"""
Integration and UI tests
"""

# Standard Library
import json
from http import HTTPStatus
from unittest.mock import patch

# Third Party
import requests
from django_webtest import WebTest
from faker import Faker

# Django
from django.contrib.messages import get_messages
from django.urls import reverse

# AA Forum
from aa_forum.helper.text import string_cleanup
from aa_forum.models import (
    Board,
    Category,
    Message,
    PersonalMessage,
    Setting,
    Topic,
    UserProfile,
)
from aa_forum.tests.utils import (
    create_fake_message,
    create_fake_messages,
    create_fake_user,
)

fake = Faker()


class TestForumUI(WebTest):
    """
    Tests for the Forum UI
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.user_1001 = create_fake_user(
            1001, "Bruce Wayne", permissions=["aa_forum.basic_access"]
        )
        cls.user_1002 = create_fake_user(
            1002, "Peter Parker", permissions=["aa_forum.basic_access"]
        )
        cls.user_1003 = create_fake_user(1003, "Lex Luthor", permissions=[])

        cls.category = Category.objects.create(name="Science")
        cls.board = Board.objects.create(name="Physics", category=cls.category)
        cls.board_with_webhook = Board.objects.create(
            name="Chemistry",
            category=cls.category,
            discord_webhook="https://discord.com/webhook/",
        )
        cls.board_with_webhook_for_replies = Board.objects.create(
            name="Biology",
            category=cls.category,
            discord_webhook="https://discord.com/webhook/",
            use_webhook_for_replies=True,
        )

    def test_should_show_forum_index(self):
        """
        Test should show forum index
        :return:
        """

        # given
        self.app.set_user(self.user_1001)

        # when
        response = self.app.get(reverse("aa_forum:forum_index"))

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(response, "aa_forum/view/forum/index.html")

    def test_should_not_show_forum_index(self):
        """
        Test should not show forum index
        :return:
        """

        # given
        self.app.set_user(self.user_1003)

        # when
        response = self.app.get(reverse("aa_forum:forum_index"))

        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(response, "/account/login/?next=/forum/")

    def test_should_create_new_topic(self):
        """
        Test should create new topic
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")
        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form["message"] = "Energy of the Higgs boson"
        page = form.submit().follow()

        # then
        self.assertEqual(self.board.topics.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_topic = Topic.objects.last()
        self.assertEqual(new_topic.subject, "Recent Discoveries")
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, "Energy of the Higgs boson")

    def test_should_return_cleaned_message_string_on_topic_creation(self):
        """
        Test should return a clean/sanitized message string when new topic is created
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board.get_absolute_url())
        dirty_message = (
            'this is a script test. <script type="text/javascript">alert('
            "'test')</script>and this is style test. <style>.MathJax, "
            ".MathJax_Message, .MathJax_Preview{display: none}</style>end tests."
        )
        cleaned_message = string_cleanup(dirty_message)

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")
        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Message Cleanup Test"
        form["message"] = dirty_message
        page = form.submit().follow()

        # then
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, cleaned_message)

    def test_should_not_create_new_topic_doe_to_subject_missing(self):
        """
        Test should not create new topic due to missing/empty subject
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")
        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = ""  # Omitting mandatory field
        form["message"] = "Energy of the Higgs boson"
        page = form.submit()

        # then
        self.assertEqual(self.board.topics.count(), 0)  # No topic created
        self.assertTemplateUsed(page, "aa_forum/view/forum/new-topic.html")

        expected_message = (
            "<h4>Error!</h4>"
            "<p>Either subject or message is missing. "
            "Please make sure you enter both fields, "
            "as both fields are mandatory.</p>"
        )
        messages = list(page.context["messages"])
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_not_create_new_topic_doe_to_message_missing(self):
        """
        Test should not create new topic due to missing/empty message
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")
        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form["message"] = ""  # Omitting mandatory field
        page = form.submit()

        # then
        self.assertEqual(self.board.topics.count(), 0)  # No topic created
        self.assertTemplateUsed(page, "aa_forum/view/forum/new-topic.html")

        expected_message = (
            "<h4>Error!</h4>"
            "<p>Either subject or message is missing. "
            "Please make sure you enter both fields, "
            "as both fields are mandatory.</p>"
        )
        messages = list(page.context["messages"])
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_not_create_topic_that_already_exists(self):
        """
        Test should not re-create an existing topic
        :return:
        """

        # given
        board = Board.objects.create(name="Space", category=self.category)
        existing_topic = Topic.objects.create(subject="Mysteries", board=board)
        create_fake_messages(existing_topic, 15)

        existing_topic_url = reverse(
            "aa_forum:forum_topic",
            kwargs={
                "category_slug": existing_topic.board.category.slug,
                "board_slug": existing_topic.board.slug,
                "topic_slug": existing_topic.slug,
            },
        )

        self.app.set_user(self.user_1001)
        page = self.app.get(board.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")
        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Mysteries"
        form["message"] = "Energy of the Higgs boson"
        page = form.submit()

        # then
        self.assertEqual(self.board.topics.count(), 0)
        self.assertTemplateUsed(page, "aa_forum/view/forum/new-topic.html")

        messages = list(page.context["messages"])
        expected_message = (
            "<h4>Warning!</h4>"
            "<p>There is already a topic with the exact same "
            "subject in this board.</p><p>See here: "
            f'<a href="{existing_topic_url}">'
            f"{existing_topic.subject}</a>"
            "</p>"
        )

        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    @patch("requests.post")
    def test_should_post_to_webhook_on_create_reply_in_topic(self, mock_post):
        """
        Test should post to Discord webhook when reply in topic
        :return:
        """

        # given
        topic = Topic.objects.create(
            subject="Mysteries", board=self.board_with_webhook_for_replies
        )
        create_fake_messages(topic=topic, amount=5)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())

        # when
        form = page.forms["aa-forum-form-message-reply"]
        form["message"] = "What is dark matter?"
        page = form.submit().follow().follow()

        # then
        self.assertEqual(topic.messages.count(), 6)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, "What is dark matter?")

        info = {"test1": "value1", "test2": "value2"}
        requests.post(self.board_with_webhook.discord_webhook, data=json.dumps(info))
        mock_post.assert_called_with(
            self.board_with_webhook.discord_webhook, data=json.dumps(info)
        )

    @patch("requests.post")
    def test_should_post_to_discord_webhook_on_create_new_topic(self, mock_post):
        """
        Test should post to Discord webhook when new topic is created
        :param mock_post:
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board_with_webhook.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")

        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form["message"] = "Energy of the Higgs boson"
        page = form.submit().follow()

        # then
        self.assertEqual(self.board_with_webhook.topics.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_topic = Topic.objects.last()
        self.assertEqual(new_topic.subject, "Recent Discoveries")
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, "Energy of the Higgs boson")

        info = {"test1": "value1", "test2": "value2"}
        requests.post(self.board_with_webhook.discord_webhook, data=json.dumps(info))
        mock_post.assert_called_with(
            self.board_with_webhook.discord_webhook, data=json.dumps(info)
        )

    @patch("requests.post")
    def test_should_post_to_discord_webhook_with_image_on_create_new_topic(
        self, mock_post
    ):
        """
        Test should post to Discord webhook when new topic with image is created
        :param mock_post:
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board_with_webhook.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")

        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form[
            "message"
        ] = "Energy of the Higgs boson <img src='/images/images/038/929/227/large/marc-bell-2a.jpg'>"  # pylint: disable=line-too-long
        page = form.submit().follow()

        # then
        self.assertEqual(self.board_with_webhook.topics.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")

        new_topic = Topic.objects.last()
        self.assertEqual(new_topic.subject, "Recent Discoveries")

        new_message = Message.objects.last()
        self.assertEqual(
            new_message.message,
            "Energy of the Higgs boson <img src='/images/images/038/929/227/large/marc-bell-2a.jpg'>",  # pylint: disable=line-too-long
        )

        info = {"test1": "value1", "test2": "value2"}
        requests.post(self.board_with_webhook.discord_webhook, data=json.dumps(info))
        mock_post.assert_called_with(
            self.board_with_webhook.discord_webhook, data=json.dumps(info)
        )

    @patch("requests.post")
    def test_should_post_to_discord_webhook_with_image_with_full_url_on_create_new_topic(  # pylint: disable=line-too-long
        self, mock_post
    ):
        """
        Test should post to Discord webhook when new topic
        with image (full url) is created
        :param mock_post:
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board_with_webhook.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")

        form = page.forms["aa-forum-form-new-topic"]
        form["subject"] = "Recent Discoveries"
        form[
            "message"
        ] = "Energy of the Higgs boson <img src='https://cdnb.artstation.com/p/assets/images/images/038/929/227/large/marc-bell-2a.jpg'>"  # pylint: disable=line-too-long
        page = form.submit().follow()

        # then
        self.assertEqual(self.board_with_webhook.topics.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_topic = Topic.objects.last()
        self.assertEqual(new_topic.subject, "Recent Discoveries")
        new_message = Message.objects.last()
        self.assertEqual(
            new_message.message,
            "Energy of the Higgs boson <img src='https://cdnb.artstation.com/p/assets/images/images/038/929/227/large/marc-bell-2a.jpg'>",  # pylint: disable=line-too-long
        )

        info = {"test1": "value1", "test2": "value2"}
        requests.post(self.board_with_webhook.discord_webhook, data=json.dumps(info))
        mock_post.assert_called_with(
            self.board_with_webhook.discord_webhook, data=json.dumps(info)
        )

    def test_should_cancel_new_topic(self):
        """
        Test should cancel new topic creation
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(self.board.get_absolute_url())

        # when
        page = page.click(linkid="aa-forum-btn-new-topic-above-list")
        page = page.click(linkid="aa-forum-btn-cancel")

        # then
        self.assertEqual(self.board.topics.count(), 0)
        self.assertTemplateUsed(page, "aa_forum/view/forum/board.html")

    def test_should_create_reply_in_topic(self):
        """
        Test should create reply in topic
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_messages(topic=topic, amount=5)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())

        # when
        form = page.forms["aa-forum-form-message-reply"]
        form["message"] = "What is dark matter?"
        page = form.submit().follow().follow()

        # then
        self.assertEqual(topic.messages.count(), 6)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, "What is dark matter?")

    def test_should_return_cleaned_message_string_on_topic_reply(self):
        """
        Test should return a clean/sanitized message string on reply in topic
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_messages(topic=topic, amount=5)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())
        dirty_message = (
            'this is a script test. <script type="text/javascript">alert('
            "'test')</script>and this is style test. <style>.MathJax, "
            ".MathJax_Message, .MathJax_Preview{display: none}</style>end tests."
        )
        cleaned_message = string_cleanup(dirty_message)

        # when
        form = page.forms["aa-forum-form-message-reply"]
        form["message"] = dirty_message
        page = form.submit().follow().follow()

        # then
        new_message = Message.objects.last()
        self.assertEqual(new_message.message, cleaned_message)

    def test_should_not_create_reply_in_topic_due_to_missing_message(self):
        """
        Test should not create reply in topic, because message field is empty
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        topic_messages = create_fake_messages(topic=topic, amount=5)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())

        # when
        form = page.forms["aa-forum-form-message-reply"]
        form["message"] = ""  # Omit mandatory field
        page = form.submit().follow()

        # then
        self.assertEqual(topic.messages.count(), 5)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        self.assertEqual(topic.last_message.message, topic_messages[-1].message)

        expected_message = (
            "<h4>Error!</h4><p>Message field is mandatory and cannot be empty.</p>"
        )
        messages = list(page.context["messages"])
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_trigger_error_message_when_trying_to_access_message_reply_directly(
        self,
    ):
        """
        Test should trigger an error message when trying to access
        the reply endpoint of a topic directly
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_messages(topic=topic, amount=5)
        self.client.force_login(self.user_1001)

        # when
        response = self.client.get(
            reverse(
                "aa_forum:forum_topic_reply",
                args=[topic.board.category.slug, topic.board.slug, topic.slug],
            ),
        )

        # then
        expected_message = (
            "<h4>Error!</h4><p>Something went wrong, please try again.</p>"
        )
        messages = list(get_messages(response.wsgi_request))
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_update_own_message(self):
        """
        Test should update own message
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        own_message = create_fake_message(topic=topic, user=self.user_1001)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())

        # when
        page = page.click(linkid=f"aa-forum-btn-modify-message-{own_message.pk}")
        form = page.forms["aa-forum-form-message-modify"]
        form["message"] = "What is dark matter?"
        page = form.submit().follow().follow()

        # then
        self.assertEqual(topic.messages.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        own_message.refresh_from_db()
        self.assertEqual(own_message.message, "What is dark matter?")

    def test_should_return_cleaned_message_string_on_update_own_message(self):
        """
        Test should return a clean/sanitized message string when updating own message
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        own_message = create_fake_message(topic=topic, user=self.user_1001)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())
        dirty_message = (
            'this is a script test. <script type="text/javascript">alert('
            "'test')</script>and this is style test. <style>.MathJax, "
            ".MathJax_Message, .MathJax_Preview{display: none}</style>end tests."
        )
        cleaned_message = string_cleanup(dirty_message)

        # when
        page = page.click(linkid=f"aa-forum-btn-modify-message-{own_message.pk}")
        form = page.forms["aa-forum-form-message-modify"]
        form["message"] = dirty_message
        page = form.submit().follow().follow()

        # then
        own_message.refresh_from_db()
        self.assertEqual(own_message.message, cleaned_message)

    def test_should_trigger_error_on_message_edit_due_to_invalid_form_data(self):
        """
        Test should trigger an error message when updating a message
        due to invalid form data
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        own_message = create_fake_message(topic=topic, user=self.user_1001)
        self.app.set_user(self.user_1001)
        page = self.app.get(topic.get_absolute_url())

        # when
        page = page.click(linkid=f"aa-forum-btn-modify-message-{own_message.pk}")
        form = page.forms["aa-forum-form-message-modify"]
        form["message"] = ""  # Omit mandatory field
        page = form.submit()

        # then
        self.assertEqual(topic.messages.count(), 1)
        self.assertTemplateUsed(page, "aa_forum/view/forum/modify-message.html")

        expected_message = "<h4>Error!</h4><p>Mandatory form field is empty.</p>"
        messages = list(page.context["messages"])
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_not_be_able_to_edit_messages_from_others(self):
        """
        Test should not be able to edit messages from others
        :return:
        """

        # given
        topic = Topic.objects.create(subject="Mysteries", board=self.board)
        alien_message = create_fake_message(topic=topic, user=self.user_1002)
        self.app.set_user(self.user_1001)

        # when
        page = self.app.get(topic.get_absolute_url())

        # then
        self.assertTemplateUsed(page, "aa_forum/view/forum/topic.html")
        self.assertNotContains(page, f"aa-forum-btn-modify-message-{alien_message.pk}")

    def test_should_find_message_by_key_word(self):
        """
        Test should find message by keyword
        :return:
        """

        # given
        topic_1 = Topic.objects.create(subject="Topic 1", board=self.board)
        create_fake_messages(topic=topic_1, amount=5)
        topic_2 = Topic.objects.create(subject="Topic 2", board=self.board)
        create_fake_messages(topic=topic_2, amount=5)
        message = Message.objects.create(
            topic=topic_1, user_created=self.user_1001, message="xyz dummy123 abc"
        )
        self.app.set_user(self.user_1001)
        page = self.app.get(reverse("aa_forum:forum_index"))

        # when
        form = page.forms["aa-forum-form-search-menu"]
        form["q"] = "dummy123"
        res = form.submit()

        # then
        self.assertTemplateUsed(res, "aa_forum/view/search/results.html")
        self.assertContains(
            res,
            reverse(
                "aa_forum:forum_message",
                args=[
                    message.topic.board.category.slug,
                    message.topic.board.slug,
                    message.topic.slug,
                    message.pk,
                ],
            ),
        )


class TestAdminCategoriesAndBoardsUI(WebTest):
    """
    Tests for the Admin UI
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = create_fake_user(
            1001,
            "Bruce Wayne",
            permissions=["aa_forum.basic_access", "aa_forum.manage_forum"],
        )

    def test_should_create_category(self):
        """
        Test should create a category
        :return:
        """

        # given
        self.app.set_user(self.user)
        page = self.app.get(reverse("aa_forum:admin_categories_and_boards"))

        # when
        form = page.forms["aa-forum-form-admin-new-category"]
        form["new-category-name"] = "Category"
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/administration/categories-and-boards.html"
        )
        new_category = Category.objects.last()
        self.assertEqual(new_category.name, "Category")

    def test_should_edit_category(self):
        """
        Test should edit category
        :return:
        """

        # given
        category = Category.objects.create(name="Category")
        self.app.set_user(self.user)
        page = self.app.get(reverse("aa_forum:admin_categories_and_boards"))

        # when
        form = page.forms[f"aa-forum-form-admin-edit-category-{category.pk}"]
        form[f"edit-category-{category.pk}-name"] = "Dummy"
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/administration/categories-and-boards.html"
        )
        category.refresh_from_db()
        self.assertEqual(category.name, "Dummy")

    def test_should_add_board_to_category(self):
        """
        Test should add board to category
        :return:
        """

        # given
        category = Category.objects.create(name="Category")
        self.app.set_user(self.user)
        page = self.app.get(reverse("aa_forum:admin_categories_and_boards"))

        # when
        form = page.forms[f"aa-forum-form-admin-add-board-{category.id}"]
        form[f"new-board-in-category-{category.id}-name"] = "Board"
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/administration/categories-and-boards.html"
        )
        new_board = category.boards.last()
        self.assertEqual(new_board.name, "Board")

    def test_should_edit_board(self):
        """
        Test should edit board
        :return:
        """

        # given
        category = Category.objects.create(name="Category")
        board = Board.objects.create(name="Board", category=category)
        self.app.set_user(self.user)
        page = self.app.get(reverse("aa_forum:admin_categories_and_boards"))

        # when
        form = page.forms[f"aa-forum-form-edit-board-{board.pk}"]
        form[f"edit-board-{board.pk}-name"] = "Dummy"
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/administration/categories-and-boards.html"
        )
        board.refresh_from_db()
        self.assertEqual(board.name, "Dummy")


class TestProfileUI(WebTest):
    """
    Tests for the Forum UI
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.user_1001 = create_fake_user(
            1001, "Bruce Wayne", permissions=["aa_forum.basic_access"]
        )
        cls.user_1002 = create_fake_user(
            1002, "Peter Parker", permissions=["aa_forum.basic_access"]
        )
        cls.user_1003 = create_fake_user(1003, "Lex Luthor", permissions=[])

    def test_should_show_profile_index(self):
        """
        Test should show profile index
        :return:
        """

        # given
        self.app.set_user(self.user_1001)

        # when
        page = self.app.get(reverse("aa_forum:profile_index"))

        # then
        self.assertTemplateUsed(page, "aa_forum/view/profile/index.html")

    def test_should_not_show_profile_index(self):
        """
        Test should not show profile index
        :return:
        """

        # given
        self.app.set_user(self.user_1003)

        # when
        page = self.app.get(reverse("aa_forum:profile_index"))

        # then
        self.assertRedirects(page, "/account/login/?next=/forum/-/profile/")

    def test_should_create_user_profile(self):
        """
        Test should create a user profile, since they are only
        created when a user opens their profile page
        :return:
        """

        # given (Should raise a DoesNotExist exception)
        with self.assertRaises(UserProfile.DoesNotExist):
            UserProfile.objects.get(pk=self.user_1002.pk)

        # when (User loggs in and opens the profile page)
        self.app.set_user(self.user_1002)
        page = self.app.get(reverse("aa_forum:profile_index"))

        # then (Right template should be loaded and UserProfile object created)
        self.assertTemplateUsed(page, "aa_forum/view/profile/index.html")

        user_profile = UserProfile.objects.get(pk=self.user_1002.pk)
        self.assertEqual(user_profile.pk, self.user_1002.pk)

    def test_should_update_user_profile(self):
        """
        Test should update the user profile
        :return:
        """

        # given
        self.app.set_user(self.user_1002)
        page = self.app.get(reverse("aa_forum:profile_index"))
        user_profile = UserProfile.objects.get(pk=self.user_1002.pk)

        # when
        form = page.forms["aa-forum-form-userprofile-modify"]
        form["signature"] = "What is dark matter?"
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(page, "aa_forum/view/profile/index.html")

        user_profile_updated = UserProfile.objects.get(pk=self.user_1002.pk)

        self.assertEqual(user_profile.signature, "")
        self.assertEqual(user_profile_updated.signature, "What is dark matter?")

    def test_should_throw_error_for_too_long_signature(self):
        """
        Test should throw an error because the signature is too long
        :return:
        """

        # given
        max_signature_length = Setting.objects.get_setting(
            setting_key=Setting.USERSIGNATURELENGTH
        )
        signature = fake.text(max_signature_length * 2)

        self.app.set_user(self.user_1002)
        page = self.app.get(reverse("aa_forum:profile_index"))

        # when
        form = page.forms["aa-forum-form-userprofile-modify"]
        form["signature"] = signature

        # then
        self.assertEqual(max_signature_length, 750)
        self.assertGreater(len(signature), max_signature_length)

        page = form.submit()
        messages = list(page.context["messages"])

        expected_message = (
            "<h4>Error!</h4><p>Something went wrong, please check your input.<p>"
        )
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_throw_error_for_invalid_url(self):
        """
        Test should throw an error because the url is not valid
        :return:
        """

        # given
        self.app.set_user(self.user_1002)
        page = self.app.get(reverse("aa_forum:profile_index"))

        # when
        form = page.forms["aa-forum-form-userprofile-modify"]
        form["website_url"] = "foobar"

        # then
        page = form.submit()
        messages = list(page.context["messages"])

        expected_message = (
            "<h4>Error!</h4><p>Something went wrong, please check your input.<p>"
        )
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_return_valid_url(self):
        """
        Test should the validated URL
        :return:
        """

        # given
        self.app.set_user(self.user_1002)
        page = self.app.get(reverse("aa_forum:profile_index"))

        # when
        form = page.forms["aa-forum-form-userprofile-modify"]
        form["website_url"] = "https://test.com"
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(page, "aa_forum/view/profile/index.html")
        user_profile_updated = UserProfile.objects.get(pk=self.user_1002.pk)
        self.assertEqual(user_profile_updated.website_url, "https://test.com")

    def test_should_return_correct_model_string(self):
        """
        Test should return the correct model string
        :return:
        """

        # given
        self.app.set_user(self.user_1002)
        self.app.get(reverse("aa_forum:profile_index"))
        user_profile = UserProfile.objects.get(pk=self.user_1002.pk)

        # then
        self.assertEqual(str(user_profile), f"Forum User Profile: {self.user_1002}")

    def test_should_set_discord_dm_on_new_personal_message_to_true(self):
        """
        Test should set discord_dm_on_new_personal_message to True
        :return:
        """

        # given
        self.app.set_user(self.user_1002)
        page = self.app.get(reverse("aa_forum:profile_index"))

        # when
        form = page.forms["aa-forum-form-userprofile-modify"]
        form["discord_dm_on_new_personal_message"] = True
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(page, "aa_forum/view/profile/index.html")

        user_profile_updated = UserProfile.objects.get(pk=self.user_1002.pk)

        self.assertTrue(user_profile_updated.discord_dm_on_new_personal_message)

    def test_should_set_discord_dm_on_new_personal_message_to_false(self):
        """
        Test should set discord_dm_on_new_personal_message to False
        :return:
        """

        # given
        self.app.set_user(self.user_1002)
        page = self.app.get(reverse("aa_forum:profile_index"))

        # when
        form = page.forms["aa-forum-form-userprofile-modify"]
        form["discord_dm_on_new_personal_message"] = ""
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(page, "aa_forum/view/profile/index.html")

        user_profile_updated = UserProfile.objects.get(pk=self.user_1002.pk)

        self.assertFalse(user_profile_updated.discord_dm_on_new_personal_message)


class TestAdminForumSettingsUI(WebTest):
    """
    Tests for the Admin UI
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user_1001 = create_fake_user(
            1001,
            "Bruce Wayne",
            permissions=["aa_forum.basic_access", "aa_forum.manage_forum"],
        )
        cls.user_1002 = create_fake_user(
            1002, "Peter Parker", permissions=["aa_forum.basic_access"]
        )

    def test_should_show_forum_settings(self):
        """
        Test should show forum settings
        :return:
        """

        # given
        self.app.set_user(self.user_1001)

        # when
        page = self.app.get(reverse("aa_forum:admin_forum_settings"))

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/administration/forum-settings.html"
        )

    def test_should_not_show_forum_settings(self):
        """
        Test should not show forum settings
        :return:
        """

        # given
        self.app.set_user(self.user_1002)

        # when
        page = self.app.get(reverse("aa_forum:admin_forum_settings"))

        # then
        self.assertRedirects(
            page, "/account/login/?next=/forum/-/admin/forum-settings/"
        )

    def test_should_update_forum_settings(self):
        """
        Test should update forum settings
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(reverse("aa_forum:admin_forum_settings"))
        forum_settings = Setting.objects.get(pk=1)

        # when
        form = page.forms["aa-forum-form-settings-modify"]
        form["user_signature_length"] = 500
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/administration/forum-settings.html"
        )

        forum_settings_updated = Setting.objects.get(pk=1)

        self.assertEqual(forum_settings.user_signature_length, 750)
        self.assertEqual(forum_settings_updated.user_signature_length, 500)

    def test_should_not_update_forum_settings_on_empty_value(self):
        """
        Test should update forum settings because of an empty value in a mandatory field
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(reverse("aa_forum:admin_forum_settings"))
        forum_settings = Setting.objects.get(pk=1)

        # when
        form = page.forms["aa-forum-form-settings-modify"]
        form["user_signature_length"] = ""
        page = form.submit()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/administration/forum-settings.html"
        )

        forum_settings_updated = Setting.objects.get(pk=1)

        self.assertEqual(
            forum_settings.user_signature_length,
            forum_settings_updated.user_signature_length,
        )

        messages = list(page.context["messages"])

        expected_message = (
            "<h4>Error!</h4><p>Something went wrong, please check your input.<p>"
        )
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_return_correct_model_string(self):
        """
        Test should return the correct model string
        :return:
        """

        # given
        forum_settings = Setting.objects.get(pk=1)

        # then
        self.assertEqual(str(forum_settings), "Forum Settings")


class TestPersonalMessageUI(WebTest):  # pylint: disable=too-many-public-methods
    """
    Tests for the Forum UI
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.user_1001 = create_fake_user(
            1001, "Bruce Wayne", permissions=["aa_forum.basic_access"]
        )
        cls.user_1002 = create_fake_user(
            1002, "Peter Parker", permissions=["aa_forum.basic_access"]
        )
        cls.user_1003 = create_fake_user(1003, "Lex Luthor", permissions=[])

    def test_should_show_messages_inbox(self):
        """
        Test should show personal messages inbox
        :return:
        """

        # given
        self.app.set_user(self.user_1001)

        # when
        page = self.app.get(reverse("aa_forum:personal_messages_inbox"))

        # then
        self.assertTemplateUsed(page, "aa_forum/view/personal-messages/inbox.html")

    def test_should_not_show_messages_inbox(self):
        """
        Test should not show personal messages inbox
        :return:
        """

        # given
        self.app.set_user(self.user_1003)

        # when
        page = self.app.get(reverse("aa_forum:personal_messages_inbox"))

        # then
        self.assertRedirects(
            page, "/account/login/?next=/forum/-/personal-messages/inbox/"
        )

    def test_should_show_messages_new_message(self):
        """
        Test should show personal messages - new message
        :return:
        """

        # given
        self.app.set_user(self.user_1001)

        # when
        page = self.app.get(reverse("aa_forum:personal_messages_new_message"))

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/personal-messages/new-message.html"
        )

    def test_should_not_show_messages_new_messages(self):
        """
        Test should not show personal messages - new message
        :return:
        """

        # given
        self.app.set_user(self.user_1003)

        # when
        page = self.app.get(reverse("aa_forum:personal_messages_new_message"))

        # then
        self.assertRedirects(
            page, "/account/login/?next=/forum/-/personal-messages/new-message/"
        )

    def test_should_show_messages_new_sent_messages(self):
        """
        Test should show personal messages - sent messages
        :return:
        """

        # given
        self.app.set_user(self.user_1001)

        # when
        page = self.app.get(reverse("aa_forum:personal_messages_sent_messages"))

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/personal-messages/sent-messages.html"
        )

    def test_should_not_show_messages_sent_messages(self):
        """
        Test should not show personal messages - sent messages
        :return:
        """

        # given
        self.app.set_user(self.user_1003)

        # when
        page = self.app.get(reverse("aa_forum:personal_messages_sent_messages"))

        # then
        self.assertRedirects(
            page, "/account/login/?next=/forum/-/personal-messages/sent-messages/"
        )

    def test_should_send_personal_message(self):
        """
        Test should send a personal message
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(reverse("aa_forum:personal_messages_new_message"))

        # when
        form = page.forms["aa-forum-form-new-personal-message"]
        form["recipient"] = self.user_1002.pk
        form["subject"] = "Foobar"
        form["message"] = "Foobar"
        page = form.submit().follow()

        # then
        self.assertTemplateUsed(page, "aa_forum/view/personal-messages/inbox.html")

        personal_message = PersonalMessage.objects.last()

        self.assertEqual(personal_message.sender, self.user_1001)
        self.assertEqual(personal_message.recipient, self.user_1002)
        self.assertEqual(personal_message.subject, "Foobar")
        self.assertEqual(personal_message.message, "Foobar")

    def test_should_not_send_personal_message_ad_rais_an_error_due_to_empty_recipient(
        self,
    ):
        """
        Test should not send a personal message and raise an error
        because of empty recipient
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(reverse("aa_forum:personal_messages_new_message"))

        # when
        form = page.forms["aa-forum-form-new-personal-message"]
        form["subject"] = "Foobar"
        form["message"] = "Foobar"
        page = form.submit()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/personal-messages/new-message.html"
        )

        messages = list(page.context["messages"])

        expected_message = (
            "<h4>Error!</h4><p>Something went wrong, please check your input.<p>"
        )
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_not_send_personal_message_ad_rais_an_error_due_to_empty_subject(
        self,
    ):
        """
        Test should not send a personal message and raise an error
        because of empty subject
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(reverse("aa_forum:personal_messages_new_message"))

        # when
        form = page.forms["aa-forum-form-new-personal-message"]
        form["recipient"] = self.user_1002.pk
        form["message"] = "Foobar"
        page = form.submit()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/personal-messages/new-message.html"
        )

        messages = list(page.context["messages"])

        expected_message = (
            "<h4>Error!</h4><p>Something went wrong, please check your input.<p>"
        )
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_not_send_personal_message_ad_rais_an_error_due_to_empty_message(
        self,
    ):
        """
        Test should not send a personal message and raise an error
        because of empty message
        :return:
        """

        # given
        self.app.set_user(self.user_1001)
        page = self.app.get(reverse("aa_forum:personal_messages_new_message"))

        # when
        form = page.forms["aa-forum-form-new-personal-message"]
        form["recipient"] = self.user_1002.pk
        form["subject"] = "Foobar"
        page = form.submit()

        # then
        self.assertTemplateUsed(
            page, "aa_forum/view/personal-messages/new-message.html"
        )

        messages = list(page.context["messages"])

        expected_message = (
            "<h4>Error!</h4><p>Something went wrong, please check your input.<p>"
        )
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_return_empty_response_for_template_for_ajax_read_message_with_get(
        self,
    ):
        """
        Test should return empty response HTTP code
        :return:
        """

        # given
        self.app.set_user(self.user_1001)

        # when
        page = self.app.get(
            reverse("aa_forum:personal_messages_ajax_read_message", args=["inbox"])
        )

        # then
        self.assertEqual(page.status_code, HTTPStatus.NO_CONTENT)

    def test_should_fail_silently_for_ajax_read_message_with_no_post_data(
        self,
    ):
        """
        Test should fail silently with no POST data and return empty response HTTP code
        :return:
        """

        # given
        self.client.force_login(self.user_1001)

        # when
        page = self.client.post(
            reverse("aa_forum:personal_messages_ajax_read_message", args=["inbox"]),
        )

        # then
        self.assertEqual(page.status_code, HTTPStatus.NO_CONTENT)

    def test_should_not_return_inbox_message_when_recipient_and_user_dont_match(
        self,
    ):
        """
        Test should not return an inbox message when recipient and user don't match
        and return empty response HTTP code
        :return:
        """

        # given
        self.client.force_login(self.user_1001)

        # when
        page = self.client.post(
            reverse("aa_forum:personal_messages_ajax_read_message", args=["inbox"]),
            data={"sender": 1, "recipient": 1, "message": 1},
        )

        # then
        self.assertEqual(page.status_code, HTTPStatus.NO_CONTENT)

    def test_should_not_return_inbox_message_when_message_doesnt_exist(
        self,
    ):
        """
        Test should not return an inbox message when recipient and user don't match
        and return empty response HTTP code
        :return:
        """

        # given
        self.client.force_login(self.user_1001)

        # when
        page = self.client.post(
            reverse("aa_forum:personal_messages_ajax_read_message", args=["inbox"]),
            data={"sender": 1, "recipient": self.user_1001.pk, "message": 1},
        )

        # then
        self.assertEqual(page.status_code, HTTPStatus.NO_CONTENT)

    def test_should_return_inbox_message(self):
        """
        Test should return an inbox message
        :return:
        """

        # given
        PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        ).save()
        message = PersonalMessage.objects.last()
        self.client.force_login(self.user_1001)

        # when
        page = self.client.post(
            reverse("aa_forum:personal_messages_ajax_read_message", args=["inbox"]),
            data={
                "sender": self.user_1002.pk,
                "recipient": self.user_1001.pk,
                "message": message.pk,
            },
        )

        # then
        self.assertEqual(page.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(
            page, "aa_forum/ajax-render/personal-messages/message.html"
        )

    def test_should_return_sent_item_message(self):
        """
        Test should return a message sent
        :return:
        """

        # given
        PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        ).save()
        message = PersonalMessage.objects.last()
        self.client.force_login(self.user_1002)

        # when
        page = self.client.post(
            reverse(
                "aa_forum:personal_messages_ajax_read_message", args=["sent-messages"]
            ),
            data={
                "sender": self.user_1002.pk,
                "recipient": self.user_1001.pk,
                "message": message.pk,
            },
        )

        # then
        self.assertEqual(page.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(
            page, "aa_forum/ajax-render/personal-messages/message.html"
        )

    def test_should_mark_message_as_read(self):
        """
        Test should unread message as read upon return
        :return:
        """

        # given
        PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        ).save()
        message_sent = PersonalMessage.objects.last()
        self.client.force_login(self.user_1001)

        # when
        page = self.client.post(
            reverse("aa_forum:personal_messages_ajax_read_message", args=["inbox"]),
            data={
                "sender": self.user_1002.pk,
                "recipient": self.user_1001.pk,
                "message": message_sent.pk,
            },
        )

        # then
        message_returned = PersonalMessage.objects.get(
            sender=self.user_1002, recipient=self.user_1001, pk=message_sent.pk
        )
        self.assertEqual(page.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(
            page, "aa_forum/ajax-render/personal-messages/message.html"
        )
        self.assertTrue(message_returned.is_read)

    def test_should_return_inbox_message_unread_count(self):
        """
        Test should return an inbox message
        :return:
        """

        # given
        PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        ).save()
        self.client.force_login(self.user_1001)

        # when
        response = self.client.post(
            reverse("aa_forum:personal_messages_ajax_unread_messages_count"),
        )

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertJSONEqual(
            str(response.content, encoding="utf8"), {"unread_messages_count": 1}
        )

    def test_should_remove_inbox_message(self):
        """
        Test should mark an inbox message as deleted_by_recipient
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        )
        message.save()

        self.client.force_login(self.user_1001)

        # when
        response = self.client.get(
            reverse(
                "aa_forum:personal_messages_message_delete", args=["inbox", message.pk]
            ),
        )

        # then
        message_removed = PersonalMessage.objects.get(pk=message.pk)
        self.assertTrue(message_removed.deleted_by_recipient)
        self.assertRedirects(
            response,
            "/forum/-/personal-messages/inbox/",
            status_code=HTTPStatus.FOUND,
        )

    def test_should_delete_inbox_message(self):
        """
        Test should delete an inbox message
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
            deleted_by_sender=True,
        )
        message.save()

        self.client.force_login(self.user_1001)

        # when
        response = self.client.get(
            reverse(
                "aa_forum:personal_messages_message_delete", args=["inbox", message.pk]
            ),
        )

        # then
        self.assertRedirects(
            response,
            "/forum/-/personal-messages/inbox/",
            status_code=HTTPStatus.FOUND,
        )

        with self.assertRaises(PersonalMessage.DoesNotExist):
            PersonalMessage.objects.get(pk=message.pk)

    def test_should_not_delete_inbox_message_and_redirect(self):
        """
        Test should not delete an inbox message and throw an error message
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
            deleted_by_sender=True,
        )
        message.save()

        self.client.force_login(self.user_1002)

        # when
        response = self.client.get(
            reverse(
                "aa_forum:personal_messages_message_delete", args=["inbox", message.pk]
            ),
        )

        # then
        self.assertTrue(PersonalMessage.objects.get(pk=message.pk))
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response,
            "/forum/-/personal-messages/inbox/",
            status_code=HTTPStatus.FOUND,
        )

    def test_should_remove_sent_message(self):
        """
        Test should mark a sent message as deleted_by_sender
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        )
        message.save()

        self.client.force_login(self.user_1002)

        # when
        response = self.client.get(
            reverse(
                "aa_forum:personal_messages_message_delete",
                args=["sent-messages", message.pk],
            ),
        )

        # then
        message_removed = PersonalMessage.objects.get(pk=message.pk)
        self.assertTrue(message_removed.deleted_by_sender)
        self.assertRedirects(
            response,
            "/forum/-/personal-messages/sent-messages/",
            status_code=HTTPStatus.FOUND,
        )

    def test_should_delete_sent_message(self):
        """
        Test should delete a sent message
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
            deleted_by_recipient=True,
        )
        message.save()

        self.client.force_login(self.user_1002)

        # when
        response = self.client.get(
            reverse(
                "aa_forum:personal_messages_message_delete",
                args=["sent-messages", message.pk],
            ),
        )

        # then
        self.assertRedirects(
            response,
            "/forum/-/personal-messages/sent-messages/",
            status_code=HTTPStatus.FOUND,
        )

        with self.assertRaises(PersonalMessage.DoesNotExist):
            PersonalMessage.objects.get(pk=message.pk)

    def test_should_not_delete_sent_message_and_redirect(self):
        """
        Test should not delete an inbox message and throw an error message
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
            deleted_by_sender=True,
        )
        message.save()

        self.client.force_login(self.user_1001)

        # when
        response = self.client.get(
            reverse(
                "aa_forum:personal_messages_message_delete",
                args=["sent-messages", message.pk],
            ),
        )

        # then
        self.assertTrue(PersonalMessage.objects.get(pk=message.pk))
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response,
            "/forum/-/personal-messages/sent-messages/",
            status_code=HTTPStatus.FOUND,
        )

    def test_should_simply_redirect_because_wrong_url_parameter(self):
        """
        Test should simply redirect to personal messages
        inbox, because wrong URL parameter
        :return:
        """

        # given
        self.client.force_login(self.user_1001)

        # when
        response = self.client.get(
            reverse(
                "aa_forum:personal_messages_message_delete",
                args=["wrong-parameter", 9999],
            ),
        )

        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response,
            "/forum/-/personal-messages/inbox/",
            status_code=HTTPStatus.FOUND,
        )

    def test_should_open_reply_view(self):
        """
        Test should show reply view
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        )
        message.save()

        # when
        self.client.force_login(self.user_1001)
        response = self.client.get(
            reverse("aa_forum:personal_messages_message_reply", args=[message.pk]),
        )

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTemplateUsed(
            response, "aa_forum/view/personal-messages/reply-message.html"
        )

    def test_should_not_open_reply_view(self):
        """
        Test not should show reply view
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        )
        message.save()

        # when
        self.client.force_login(self.user_1002)
        response = self.client.get(
            reverse("aa_forum:personal_messages_message_reply", args=[message.pk]),
        )

        # then
        self.assertEqual(response.status_code, HTTPStatus.FOUND)
        self.assertRedirects(
            response,
            "/forum/-/personal-messages/inbox/",
            status_code=HTTPStatus.FOUND,
        )

    def test_should_send_reply(self):
        """
        Test should send a reply to a personal message
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        )
        message.save()

        # when
        self.app.set_user(self.user_1001)
        page = self.app.get(
            reverse("aa_forum:personal_messages_message_reply", args=[message.pk]),
        )

        form = page.forms["aa-forum-form-new-personal-message-reply"]
        form["message"] = "BARFOO"
        response = form.submit().follow()

        # then
        reply = PersonalMessage.objects.last()
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(reply.subject, "Re: Test Message")
        self.assertEqual(reply.message, "BARFOO")
        self.assertEqual(reply.sender, self.user_1001)
        self.assertEqual(reply.recipient, self.user_1002)
        self.assertTemplateUsed(response, "aa_forum/view/personal-messages/inbox.html")

    def test_should_not_send_reply_with_missing_form_field(self):
        """
        Test should not send a reply to a personal message because form field is missing
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        )
        message.save()

        # when
        self.app.set_user(self.user_1001)
        page = self.app.get(
            reverse("aa_forum:personal_messages_message_reply", args=[message.pk]),
        )

        form = page.forms["aa-forum-form-new-personal-message-reply"]
        response = form.submit()

        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        expected_message = (
            "<h4>Error!</h4><p>Something went wrong, please check your input.<p>"
        )
        messages = list(response.context["messages"])
        self.assertEqual(len(messages), 1)
        self.assertEqual(str(messages[0]), expected_message)

    def test_should_change_topic_on_reply(self):
        """
        Test should add "Re:" to the topic on first reply
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        )
        message.save()

        # when
        self.app.set_user(self.user_1001)
        page = self.app.get(
            reverse("aa_forum:personal_messages_message_reply", args=[message.pk]),
        )

        form = page.forms["aa-forum-form-new-personal-message-reply"]
        form["message"] = "BARFOO"
        response = form.submit().follow()

        # then
        reply = PersonalMessage.objects.last()
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(reply.subject, "Re: Test Message")

    def test_should_not_change_topic_on_reply(self):
        """
        Test should not add another "Re:" to the topic on reply to a reply
        :return:
        """

        # given
        message = PersonalMessage(
            subject="Re: Test Message",
            sender=self.user_1002,
            recipient=self.user_1001,
            message="FOOBAR",
        )
        message.save()

        # when
        self.app.set_user(self.user_1001)
        page = self.app.get(
            reverse("aa_forum:personal_messages_message_reply", args=[message.pk]),
        )

        form = page.forms["aa-forum-form-new-personal-message-reply"]
        form["message"] = "BARFOO"
        response = form.submit().follow()

        # then
        reply = PersonalMessage.objects.last()
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertEqual(reply.subject, "Re: Test Message")



"""
AA-Forum path config
"""

# Django
from django.urls import include, path

# AA Forum
from aa_forum.constants import INTERNAL_URL_PREFIX
from aa_forum.views import admin, forum, personal_messages, profile, search

app_name: str = "aa_forum"

# Internal URLs
internal_urls = [
    # Admin URLs (Menu Item :: Categories and Boards)
    path(
        route="admin/categories-and-boards/",
        view=admin.categories_and_boards,
        name="admin_categories_and_boards",
    ),
    path(
        route="admin/categories-and-boards/category/create/",
        view=admin.category_create,
        name="admin_category_create",
    ),
    path(
        route="admin/categories-and-boards/category/<int:category_id>/edit/",
        view=admin.category_edit,
        name="admin_category_edit",
    ),
    path(
        route="admin/categories-and-boards/category/<int:category_id>/delete/",
        view=admin.category_delete,
        name="admin_category_delete",
    ),
    path(
        route="admin/categories-and-boards/category/<int:category_id>/board/create/",
        view=admin.board_create,
        name="admin_board_create",
    ),
    path(
        route=(
            "admin/categories-and-boards/category/"
            "<int:category_id>/board/<int:board_id>/create-child-board/"
        ),
        view=admin.board_create_child,
        name="admin_board_create_child",
    ),
    path(
        route=(
            "admin/categories-and-boards/category/"
            "<int:category_id>/board/<int:board_id>/edit/"
        ),
        view=admin.board_edit,
        name="admin_board_edit",
    ),
    path(
        route=(
            "admin/categories-and-boards/category/"
            "<int:category_id>/board/<int:board_id>/delete/"
        ),
        view=admin.board_delete,
        name="admin_board_delete",
    ),
    # Admin Ajax URLs (Menu Item :: Categories and Boards)
    path(
        route="ajax/admin/categories-and-boards/category-order/",
        view=admin.ajax_category_order,
        name="admin_ajax_category_order",
    ),
    path(
        route="ajax/admin/categories-and-boards/board-order/",
        view=admin.ajax_board_order,
        name="admin_ajax_board_order",
    ),
    # Admin URLs (Menu Item :: Forum Settings)
    path(
        route="admin/forum-settings/",
        view=admin.forum_settings,
        name="admin_forum_settings",
    ),
    # Profile URLs
    path(
        route="profile/",
        view=profile.index,
        name="profile_index",
    ),
    # Personal Messages URLs
    path(
        route="personal-messages/inbox/",
        view=personal_messages.inbox,
        name="personal_messages_inbox",
    ),
    path(
        route="personal-messages/inbox/page/<int:page_number>/",
        view=personal_messages.inbox,
        name="personal_messages_inbox",
    ),
    path(
        route="personal-messages/new-message/",
        view=personal_messages.new_message,
        name="personal_messages_new_message",
    ),
    path(
        route="personal-messages/sent-messages/",
        view=personal_messages.sent_messages,
        name="personal_messages_sent_messages",
    ),
    path(
        route="personal-messages/sent-messages/page/<int:page_number>/",
        view=personal_messages.sent_messages,
        name="personal_messages_sent_messages",
    ),
    path(
        route="personal-messages/inbox/message/<int:message_id>/reply/",
        view=personal_messages.reply_message,
        name="personal_messages_message_reply",
    ),
    path(
        route="personal-messages/<str:folder>/message/<int:message_id>/delete/",
        view=personal_messages.delete_message,
        name="personal_messages_message_delete",
    ),
    # Personal Messages Ajax URLs
    path(
        route="ajax/personal-messages/<str:folder>/read-message/",
        view=personal_messages.ajax_read_message,
        name="personal_messages_ajax_read_message",
    ),
    path(
        route="ajax/personal-messages/unread-messages-count/",
        view=personal_messages.ajax_unread_messages_count,
        name="personal_messages_ajax_unread_messages_count",
    ),
    # Service URLs
    path(
        route="message/<int:message_id>/delete/",
        view=forum.message_delete,
        name="forum_message_delete",
    ),
    path(
        route="topic/<int:topic_id>/change-lock-state/",
        view=forum.topic_change_lock_state,
        name="forum_topic_change_lock_state",
    ),
    path(
        route="topic/<int:topic_id>/change-sticky-state/",
        view=forum.topic_change_sticky_state,
        name="forum_topic_change_sticky_state",
    ),
    path(
        route="topic/<int:topic_id>/delete/",
        view=forum.topic_delete,
        name="forum_topic_delete",
    ),
    path(
        route="search/",
        view=search.results,
        name="search_results",
    ),
    path(
        route="search/page/<int:page_number>/",
        view=search.results,
        name="search_results",
    ),
    path(
        route="unread/",
        view=forum.topic_show_all_unread,
        name="forum_topic_show_all_unread",
    ),
    path(
        route="mark-all-as-read/",
        view=forum.mark_all_as_read,
        name="forum_mark_all_as_read",
    ),
]

# Forum "public" URLs
forum_urls = [
    path(route="", view=forum.index, name="forum_index"),
    # Forum URLs :: Board
    path(
        route="<slug:category_slug>/<slug:board_slug>/",
        view=forum.board,
        name="forum_board",
    ),
    path(
        route="<slug:category_slug>/<slug:board_slug>/page/<int:page_number>/",
        view=forum.board,
        name="forum_board",
    ),
    path(
        route=(
            f"{INTERNAL_URL_PREFIX}/<slug:category_slug>/<slug:board_slug>/new-topic/"
        ),
        view=forum.board_new_topic,
        name="forum_board_new_topic",
    ),
    # Forum URLs :: Topic
    path(
        route="<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/",
        view=forum.topic,
        name="forum_topic",
    ),
    path(
        route=(
            "<slug:category_slug>/<slug:board_slug>/"
            "<slug:topic_slug>/page/<int:page_number>/"
        ),
        view=forum.topic,
        name="forum_topic",
    ),
    path(
        route="<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/reply/",
        view=forum.topic_reply,
        name="forum_topic_reply",
    ),
    path(
        route="<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/unread/",
        view=forum.topic_first_unread_message,
        name="forum_topic_first_unread_message",
    ),
    path(
        route="<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/modify/",
        view=forum.topic_modify,
        name="forum_topic_modify",
    ),
    # Forum URLs :: Message
    path(
        route=(
            "<slug:category_slug>/<slug:board_slug>/<slug:topic_slug>/<int:message_id>/"
        ),
        view=forum.message,
        name="forum_message",
    ),
    path(
        route=(
            "<slug:category_slug>/<slug:board_slug>/"
            "<slug:topic_slug>/<int:message_id>/modify/"
        ),
        view=forum.message_modify,
        name="forum_message_modify",
    ),
]

# Put it all together
# IMPORTANT
# All internal URLs need to start with the designated prefix `{INTERNAL_URL_PREFIX}` to
# prevent conflicts with user generated forum URLs
urlpatterns = [
    # Forum internal URLs (Need to be first in line)
    path(f"{INTERNAL_URL_PREFIX}/", include(internal_urls)),
    # Forum "public" URLs
    path("", include(forum_urls)),
]



"""
Test auth_hooks
"""

# Standard Library
from http import HTTPStatus

# Django
from django.test import TestCase
from django.urls import reverse

# AA Forum
from aa_forum.tests.utils import create_fake_user


class TestHooks(TestCase):
    """
    Test the app hook into allianceauth
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up groups and users
        """

        super().setUpClass()

        # User cannot access
        cls.user_1001 = create_fake_user(1001, "Peter Parker")

        # User can access
        cls.user_1002 = create_fake_user(
            1002, "Bruce Wayne", permissions=["aa_forum.basic_access"]
        )

        cls.html_menu = f"""
            <li>
                <a class href="{reverse('aa_forum:forum_index')}">
                    <i class="fas fa-comments fa-fw"></i>
                    Forum
                </a>
            </li>
        """

    def test_render_hook_success(self):
        """
        Test should show the link to the app in the navigation to user with access
        :return:
        :rtype:
        """

        self.client.force_login(self.user_1002)

        response = self.client.get(reverse("authentication:dashboard"))

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertContains(response, self.html_menu, html=True)

    def test_render_hook_fail(self):
        """
        Test should not show the link to the app in the
        navigation to user without access
        :return:
        :rtype:
        """

        self.client.force_login(self.user_1001)

        response = self.client.get(reverse("authentication:dashboard"))

        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertNotContains(response, self.html_menu, html=True)



"""
Test our signals
"""

# Django
from django.contrib.auth.models import Group
from django.test import TestCase

# AA Forum
from aa_forum.models import Board, Category
from aa_forum.tests.utils import create_fake_user

MODELS_PATH = "aa_forum.models"


class TestBoard(TestCase):
    """
    Test Board signals
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = create_fake_user(1001, "Bruce Wayne")
        cls.group = Group.objects.create(name="Superhero")

    def setUp(self) -> None:
        self.category = Category.objects.create(name="Science")

    def test_should_set_parent_board_access_restriction(self):
        """
        Test that a child board inherits the groups from its parent board on creation
        :return:
        """

        board = Board.objects.create(name="Physics", category=self.category)
        board.groups.add(self.group)
        board_2 = Board.objects.create(
            name="Thermal Theories", category=self.category, parent_board=board
        )

        self.assertEqual(list(board_2.groups.all()), [self.group])

    def test_should_set_child_board_access_restriction(self):
        """
        Test that a child parent board passes down its group restriction to
        its child on save
        :return:
        """

        board = Board.objects.create(name="Physics", category=self.category)
        board_2 = Board.objects.create(
            name="Thermal Theories", category=self.category, parent_board=board
        )
        board.groups.add(self.group)
        board.save()

        # result_groups = [groups for group in board_2.get_groups()]

        self.assertEqual(list(board_2.groups.all()), [self.group])



# Django
from django.db import migrations

default_settings_to_migrate = [
    {"variable": "defaultMaxMessages", "value": "15"},
    {"variable": "defaultMaxTopics", "value": "20"},
]


def on_migrate(apps, schema_editor):
    """
    Create default settings on migration
    :param apps:
    :param schema_editor:
    :return:
    """

    Setting = apps.get_model("aa_forum", "Setting")
    db_alias = schema_editor.connection.alias

    Setting.objects.using(db_alias).create(pk=1)


def on_migrate_zero(apps, schema_editor):
    """
    Remove default settings on migratio to zero
    :param apps:
    :param schema_editor:
    :return:
    """

    Setting = apps.get_model("aa_forum", "Setting")
    db_alias = schema_editor.connection.alias
    Setting.objects.using(db_alias).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("aa_forum", "0007_change_settings_to_singleton"),
    ]

    operations = [migrations.RunPython(on_migrate, on_migrate_zero)]



# Generated by Django 4.0.7 on 2022-08-26 00:49

# Django
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

# ckEditor
import ckeditor_uploader.fields


class Migration(migrations.Migration):

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("aa_forum", "0010_better_setting_names"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserProfile",
            fields=[
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        related_name="aa_forum_user_profile",
                        serialize=False,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "signature",
                    ckeditor_uploader.fields.RichTextUploadingField(blank=True),
                ),
                ("website_title", models.CharField(blank=True, max_length=254)),
                ("website_url", models.CharField(blank=True, max_length=254)),
            ],
            options={
                "verbose_name": "user profile",
                "verbose_name_plural": "user profiles",
                "default_permissions": (),
            },
        ),
        migrations.AddField(
            model_name="setting",
            name="user_signature_length",
            field=models.IntegerField(
                default=750,
                help_text="Maximum length of a users signature",
                verbose_name="User signature length",
            ),
        ),
    ]



"""
User profile view
"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Forum
from aa_forum import __title__
from aa_forum.forms import UserProfileForm
from aa_forum.helper.user import get_user_profile

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("aa_forum.basic_access")
def index(request: WSGIRequest) -> HttpResponse:
    """
    Profile indes view
    :param request:
    :return:
    """

    logger.info(f"{request.user} called their user profile.")

    user_profile = get_user_profile(user=request.user)

    # If this is a POST request we need to process the form data
    if request.method == "POST":
        user_profile_form = UserProfileForm(request.POST, instance=user_profile)

        # Check whether it's valid:
        if user_profile_form.is_valid():
            user_profile.signature = user_profile_form.cleaned_data["signature"]
            user_profile.website_title = user_profile_form.cleaned_data["website_title"]
            user_profile.website_url = user_profile_form.cleaned_data["website_url"]
            user_profile.discord_dm_on_new_personal_message = (
                user_profile_form.cleaned_data["discord_dm_on_new_personal_message"]
            )
            user_profile.save()

            messages.success(
                request, mark_safe(_("<h4>Success!</h4><p>Profile saved.<p>"))
            )

            return redirect("aa_forum:profile_index")
        else:
            messages.error(
                request,
                mark_safe(
                    _(
                        "<h4>Error!</h4>"
                        "<p>Something went wrong, please check your input.<p>"
                    )
                ),
            )
    else:
        user_profile_form = UserProfileForm(instance=user_profile)

    context = {"form": user_profile_form}

    return render(request, "aa_forum/view/profile/index.html", context)



"""
Messages views
"""

# Standard Library
from http import HTTPStatus

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.core.handlers.wsgi import WSGIRequest
from django.core.paginator import Paginator
from django.http import HttpResponse, JsonResponse
from django.shortcuts import redirect, render
from django.utils.datastructures import MultiValueDictKeyError
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Forum
from aa_forum import __title__
from aa_forum.forms import NewPersonalMessageForm, ReplyPersonalMessageForm
from aa_forum.helper.user import get_main_character_from_user
from aa_forum.models import PersonalMessage, Setting

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("aa_forum.basic_access")
def inbox(request: WSGIRequest, page_number: int = None) -> HttpResponse:
    """
    Messages overview
    :param request:
    :param page_number:
    :return:
    """

    logger.info(f"{request.user} called their messages overview.")

    personal_messages = PersonalMessage.objects.get_personal_messages_for_user(
        request.user
    )

    paginator = Paginator(
        personal_messages,
        int(Setting.objects.get_setting(setting_key=Setting.MESSAGESPERPAGE)),
    )
    page_obj = paginator.get_page(page_number)

    context = {"page_obj": page_obj}

    return render(request, "aa_forum/view/personal-messages/inbox.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def new_message(request: WSGIRequest) -> HttpResponse:
    """
    Create a new personal message
    :param request:
    :return:
    """

    logger.info(f"{request.user} called the new personal message page.")

    # If this is a POST request we need to process the form data
    if request.method == "POST":
        new_private_message_form = NewPersonalMessageForm(request.POST)

        # Check whether it's valid:
        if new_private_message_form.is_valid():
            sender = request.user
            recipient = new_private_message_form.cleaned_data["recipient"]
            subject = new_private_message_form.cleaned_data["subject"]
            message = new_private_message_form.cleaned_data["message"]

            PersonalMessage(
                sender=sender,
                recipient=recipient,
                subject=subject,
                message=message,
            ).save()

            recipient_main_char = get_main_character_from_user(recipient)
            messages.success(
                request,
                mark_safe(
                    _(f"<h4>Success!</h4><p>Message to {recipient_main_char} sent.<p>")
                ),
            )

            return redirect("aa_forum:personal_messages_inbox")

        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4>"
                    "<p>Something went wrong, please check your input.<p>"
                )
            ),
        )
    else:
        new_private_message_form = NewPersonalMessageForm()

    context = {"form": new_private_message_form}

    return render(request, "aa_forum/view/personal-messages/new-message.html", context)


@login_required
@permission_required("aa_forum.basic_access")
def sent_messages(request: WSGIRequest, page_number: int = None) -> HttpResponse:
    """
    Overview of all messages sent by a user
    :param request:
    :param page_number:
    :return:
    """

    logger.info(f"{request.user} called the their sent personal message page.")

    personal_messages = PersonalMessage.objects.get_personal_messages_sent_for_user(
        request.user
    )

    paginator = Paginator(
        personal_messages,
        int(Setting.objects.get_setting(setting_key=Setting.MESSAGESPERPAGE)),
    )
    page_obj = paginator.get_page(page_number)

    context = {"page_obj": page_obj}

    return render(
        request, "aa_forum/view/personal-messages/sent-messages.html", context
    )


@login_required
@permission_required("aa_forum.basic_access")
def reply_message(request: WSGIRequest, message_id: int) -> HttpResponse:
    """
    Reply to a message
    :param request:
    :param message_id:
    :return:
    """

    context = {}

    try:
        personal_message = PersonalMessage.objects.get(
            pk=message_id, recipient=request.user, deleted_by_recipient=False
        )
    except PersonalMessage.DoesNotExist:
        messages.error(
            request,
            mark_safe(
                _(
                    "<h4>Error!</h4>"
                    "<p>The message you were trying to reply to does either not exist "
                    "or you are not the recipient.<p>"
                )
            ),
        )

        return redirect("aa_forum:personal_messages_inbox")
    else:
        # If this is a POST request we need to process the form data
        if request.method == "POST":
            reply_private_message_form = ReplyPersonalMessageForm(request.POST)

            # Check whether it's valid:
            if reply_private_message_form.is_valid():
                sender = request.user
                recipient = personal_message.sender

                subject = personal_message.subject
                if not subject.startswith("Re:"):
                    subject = "Re: " + subject

                subject = subject
                message = reply_private_message_form.cleaned_data["message"]

                PersonalMessage(
                    sender=sender,
                    recipient=recipient,
                    subject=subject,
                    message=message,
                ).save()

                recipient_main_char = get_main_character_from_user(recipient)
                messages.success(
                    request,
                    mark_safe(
                        _(
                            "<h4>Success!</h4>"
                            f"<p>Reply to {recipient_main_char} sent.<p>"
                        )
                    ),
                )

                return redirect("aa_forum:personal_messages_inbox")

            messages.error(
                request,
                mark_safe(
                    _(
                        "<h4>Error!</h4>"
                        "<p>Something went wrong, please check your input.<p>"
                    )
                ),
            )
        else:
            reply_private_message_form = ReplyPersonalMessageForm()

        context["message"] = personal_message
        context["form"] = reply_private_message_form

    return render(
        request, "aa_forum/view/personal-messages/reply-message.html", context
    )


@login_required
@permission_required("aa_forum.basic_access")
def delete_message(request: WSGIRequest, folder: str, message_id: int) -> HttpResponse:
    """
    Delete a personal message
    :param request:
    :param folder:
    :param message_id:
    :return:
    """

    def folder_inbox() -> HttpResponse:
        """
        Remove message from inbox
        :return:
        """

        try:
            message = PersonalMessage.objects.get(pk=message_id, recipient=request.user)
        except PersonalMessage.DoesNotExist:
            messages.error(
                request=request,
                message=mark_safe(
                    _(
                        "<h4>Error!</h4>"
                        "<p>The message you tried to remove does either not exist "
                        "or is not yours to remove.<p>"
                    )
                ),
            )
        else:
            if message.deleted_by_sender is True:
                message.delete()
            else:
                message.deleted_by_recipient = True
                message.save()

            messages.success(
                request=request,
                message=mark_safe(
                    _("<h4>Success!</h4><p>Message removed from your inbox.<p>")
                ),
            )

        return redirect("aa_forum:personal_messages_inbox")

    def folder_sent_messages() -> HttpResponse:
        """
        Remove message from sent messages
        :return:
        """

        try:
            message = PersonalMessage.objects.get(pk=message_id, sender=request.user)
        except PersonalMessage.DoesNotExist:
            messages.error(
                request=request,
                message=mark_safe(
                    _(
                        "<h4>Error!</h4>"
                        "<p>The message you tried to remove does either not exist "
                        "or is not yours to remove.<p>"
                    )
                ),
            )
        else:
            if message.deleted_by_recipient is True:
                message.delete()
            else:
                message.deleted_by_sender = True
                message.save()

            messages.success(
                request=request,
                message=mark_safe(
                    _(
                        "<h4>Success!</h4>"
                        "<p>Message has been removed from your sent messages.<p>"
                    )
                ),
            )

        return redirect("aa_forum:personal_messages_sent_messages")

    switch = {"inbox": folder_inbox, "sent-messages": folder_sent_messages}

    if folder in switch:
        return switch[folder]()

    messages.error(
        request=request,
        message=mark_safe(_("<h4>Error!</h4><p>Something went wrong.<p>")),
    )

    return redirect("aa_forum:personal_messages_inbox")


@login_required
@permission_required("aa_forum.basic_access")
def ajax_read_message(request: WSGIRequest, folder: str) -> HttpResponse:
    """
    Ajax :: Read a personal message
    :param request:
    :param folder:
    :return:
    """

    data = {}

    if request.method == "POST":
        try:
            sender_id = int(request.POST["sender"])
            recipient_id = int(request.POST["recipient"])
            message_id = int(request.POST["message"])
        except MultiValueDictKeyError:
            # Fail silently
            pass
        else:
            if (folder == "inbox" and request.user.id == recipient_id) or (
                folder == "sent-messages" and request.user.id == sender_id
            ):
                try:
                    message = PersonalMessage.objects.get(
                        pk=message_id, sender_id=sender_id, recipient_id=recipient_id
                    )
                except PersonalMessage.DoesNotExist:
                    # Fail silently
                    pass
                else:
                    # Mark message as read
                    if folder == "inbox" and message.is_read is False:
                        message.is_read = True
                        message.save()

                    data["message"] = message
                    data["folder"] = folder

                    return render(
                        request,
                        "aa_forum/ajax-render/personal-messages/message.html",
                        data,
                    )

    return HttpResponse(status=HTTPStatus.NO_CONTENT)


@login_required
@permission_required("aa_forum.basic_access")
def ajax_unread_messages_count(request: WSGIRequest) -> JsonResponse:
    """
    Get unread messages count for a user
    :param request:
    :return:
    """

    unread_messages_count = (
        PersonalMessage.objects.get_personal_message_unread_count_for_user(request.user)
    )

    data = {"unread_messages_count": unread_messages_count}

    return JsonResponse(data, safe=False)



"""
Forms
"""

# Django
from django import forms
from django.contrib.auth.models import Group, User
from django.core.exceptions import ValidationError
from django.core.validators import MaxLengthValidator, URLValidator
from django.forms import ModelForm
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# ckEditor
from ckeditor_uploader.widgets import CKEditorUploadingWidget

# AA Forum
from aa_forum.app_settings import discord_messaging_proxy_available
from aa_forum.helper.text import string_cleanup
from aa_forum.models import (
    Board,
    Category,
    General,
    Message,
    PersonalMessage,
    Setting,
    Topic,
    UserProfile,
)


def get_mandatory_form_label_text(text):
    """
    Label text for mandatory form fields
    :param text:
    :return:
    """

    required_text = _("This field is mandatory")
    required_marker = (
        f'<span aria-label="{required_text}" class="form-required-marker">*</span>'
    )

    return mark_safe(
        f'<span class="form-field-required">{text} {required_marker}</span>'
    )


class SpecialModelChoiceIterator(forms.models.ModelChoiceIterator):
    """
    Variant of Django's ModelChoiceIterator to prevent it from always re-fetching the
    given queryset from database.
    """

    def __iter__(self):
        if self.field.empty_label is not None:
            yield "", self.field.empty_label

        queryset = self.queryset

        for obj in queryset:
            yield self.choice(obj)


class SpecialModelMultipleChoiceField(forms.ModelMultipleChoiceField):
    """
    Variant of Django's ModelMultipleChoiceField to prevent it from always
    re-fetching the given queryset from database.
    """

    iterator = SpecialModelChoiceIterator

    def _get_queryset(self):
        return self._queryset

    def _set_queryset(self, queryset):
        self._queryset = queryset
        self.widget.choices = self.choices

    queryset = property(_get_queryset, _set_queryset)


class NewTopicForm(forms.Form):
    """
    New topic form
    """

    subject = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Subject")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Subject")}),
    )

    message = forms.CharField(
        widget=CKEditorUploadingWidget(
            config_name="aa_forum",
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"},
        ),
        required=True,
        label=get_mandatory_form_label_text(_("Message")),
    )

    def clean_message(self):
        """
        Cleanup the message
        :return:
        """

        message = string_cleanup(self.cleaned_data["message"])

        return message


class EditTopicForm(ModelForm):
    """
    Edit category form
    """

    subject = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Subject")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Subject")}),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Topic
        fields = ["subject"]


class NewCategoryForm(ModelForm):
    """
    New category form
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Name")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Category Name")}),
    )
    boards = forms.CharField(
        required=False,
        label=_("Boards"),
        help_text=_(
            "Boards to be created with this category (One board per line). These "
            "boards will have no group restrictions, so you have to add them later "
            "where needed."
        ),
        widget=forms.Textarea(
            attrs={
                "rows": 10,
                "cols": 20,
                "input_type": "textarea",
                "placeholder": _("Boards"),
            }
        ),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Category
        fields = ["name", "boards"]


class EditCategoryForm(ModelForm):
    """
    Edit category form
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Name")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Category Name")}),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Category
        fields = ["name"]


class EditBoardForm(ModelForm):
    """
    Edit board form
    """

    name = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Name")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Board Name")}),
    )
    description = forms.CharField(
        required=False,
        label=_("Description"),
        widget=forms.Textarea(
            attrs={
                "rows": 10,
                "cols": 20,
                "input_type": "textarea",
                "placeholder": _("Board Description (Optional)"),
            }
        ),
    )
    groups = SpecialModelMultipleChoiceField(
        required=False,
        label=_("Group Restrictions"),
        help_text=_(
            "This will restrict access to this board to the selected groups. If no "
            "groups are selected, everyone who can access the forum can also access "
            "this board. (This setting is optional)"
        ),
        queryset=Group.objects.all(),
    )
    discord_webhook = forms.CharField(
        required=False,
        label=_("Discord Webhook (Optional)"),
        help_text=_(
            "Discord Webhook URL for the channel to post about new topics in this "
            "board. (This setting is optional) "
        ),
        max_length=254,
        widget=forms.TextInput(
            attrs={
                "placeholder": "https://discord.com/api/webhooks/xxxxxxxxxxxxxxxxxx/xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"  # pylint: disable=line-too-long
            }
        ),
    )
    use_webhook_for_replies = forms.BooleanField(
        initial=False,
        required=False,
        label=_("Use this Discord Webhook for replies as well?"),
        help_text=_(
            "When activated every reply to any topic in this board will be "
            "posted to the defined Discord channel. (Child boards are excluded) "
            "Chose wisely! (Default: NO)"
        ),
    )
    is_announcement_board = forms.BooleanField(
        initial=False,
        required=False,
        label=_("Mark Board as 'Announcement Board'"),
        help_text=_(
            "Mark this board as an 'Announcement Board', meaning that only certain "
            "selected groups can start new topics. All others who have access to this "
            "board will still be able to discuss in the topics though. This setting "
            "will not be inherited to child boards. (Default: NO)"
        ),
    )
    announcement_groups = SpecialModelMultipleChoiceField(
        required=False,
        label=_("Start Topic Restrictions for 'Announcement Boards'"),
        help_text=_(
            "User in at least one of the selected groups will be able to start topics "
            "in 'Announcement Boards'. If no group is selected, only forum admins can "
            "do so. This setting will not be inherited to child boards. (Hint: These "
            "restrictions only take effect when a board is marked as 'Announcement "
            "Board', see checkbox above.)"
        ),
        queryset=Group.objects.all(),
    )

    def __init__(self, *args, **kwargs):
        """
        When form is initialized
        :param args:
        :param kwargs:
        """

        groups_queryset = kwargs.pop("groups_queryset", None)

        super().__init__(*args, **kwargs)

        if groups_queryset:
            self.fields["groups"].queryset = groups_queryset

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Board
        fields = [
            "name",
            "description",
            "groups",
            "discord_webhook",
            "use_webhook_for_replies",
            "is_announcement_board",
            "announcement_groups",
        ]


class EditMessageForm(ModelForm):
    """
    Edit message form
    """

    message = forms.CharField(
        widget=CKEditorUploadingWidget(
            config_name="aa_forum",
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"},
        ),
        required=True,
        label=get_mandatory_form_label_text(_("Message")),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Message
        fields = ["message"]

    def clean_message(self):
        """
        Cleanup the message
        :return:
        """

        message = string_cleanup(self.cleaned_data["message"])

        return message


class UserProfileForm(ModelForm):
    """
    Edit message form
    """

    signature = forms.CharField(
        widget=CKEditorUploadingWidget(
            config_name="aa_forum",
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"},
        ),
        required=False,
        label=_("Signature"),
        help_text=_("Your signature will appear below your posts."),
    )
    website_title = forms.CharField(
        required=False,
        label=_("Website Title"),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("e.g.: My Homepage")}),
        help_text=_("Your website's title."),
    )
    website_url = forms.CharField(
        required=False,
        label=_("Website URL"),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("https://example.com")}),
        help_text=_(
            "Your website's URL. (Don't forget to also set a title for your website, "
            "otherwise this field will be ignored.)"
        ),
    )
    discord_dm_on_new_personal_message = forms.BooleanField(
        required=False,
        label=_("PM me on Discord when I get a new personal message"),
        help_text=_(
            "Information: There is currently no module installed that can handle "
            "Discord direct messages. Have a chat with your IT guys to remedy this."
        )
        if discord_messaging_proxy_available() is False
        else "",
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = UserProfile
        fields = [
            "signature",
            "website_title",
            "website_url",
            "discord_dm_on_new_personal_message",
        ]

    def clean_signature(self):
        """
        Check that the signature is not longer than allowed
        :return:
        """

        signature = string_cleanup(self.cleaned_data["signature"])

        if not signature:
            return ""

        max_signature_length = Setting.objects.get_setting(
            setting_key=Setting.USERSIGNATURELENGTH
        )

        try:
            MaxLengthValidator(max_signature_length)(signature)
        except ValidationError as exc:
            raise ValidationError(
                _(
                    f"Ensure your signature has at most {max_signature_length} "
                    f"characters. (Currently: {len(signature)})"
                )
            ) from exc
        else:
            return signature

    def clean_website_url(self):
        """
        Check if it's a valid URL
        :return:
        """

        website_url = self.cleaned_data["website_url"]

        if not website_url:
            return ""

        try:
            URLValidator()(website_url)
        except ValidationError as exc:
            raise ValidationError(_("This is not a valid URL")) from exc
        else:
            return website_url


class SettingForm(ModelForm):
    """
    Edit message form
    """

    messages_per_page = forms.IntegerField(
        required=True,
        label=get_mandatory_form_label_text(_("Messages per page")),
        help_text=_("How many messages per page should be displayed in a forum topic?"),
    )
    topics_per_page = forms.IntegerField(
        required=True,
        label=get_mandatory_form_label_text(_("Topics per page")),
        help_text=_(
            "How many topics per page should be displayed in a forum category?"
        ),
    )
    user_signature_length = forms.IntegerField(
        required=True,
        label=get_mandatory_form_label_text(_("User signature length")),
        help_text=_(
            "How long (Number of characters) is a user's signature allowed to be?"
        ),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = Setting
        fields = ["messages_per_page", "topics_per_page", "user_signature_length"]


class NewPersonalMessageForm(ModelForm):
    """
    New personal message form
    """

    recipient = forms.ModelChoiceField(
        queryset=User.objects.none(),
        required=True,
        label=get_mandatory_form_label_text(_("Recipient")),
    )
    subject = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(_("Subject")),
        max_length=254,
        widget=forms.TextInput(attrs={"placeholder": _("Hello there ...")}),
    )
    message = forms.CharField(
        widget=CKEditorUploadingWidget(
            config_name="aa_forum",
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"},
        ),
        required=True,
        label=get_mandatory_form_label_text(_("Message")),
    )

    def __init__(self, *args, **kwargs):
        """
        When form is initialized
        :param args:
        :param kwargs:
        """

        super().__init__(*args, **kwargs)

        self.fields["recipient"].queryset = General.users_with_basic_access()

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = PersonalMessage
        fields = ["recipient", "subject", "message"]

    def clean_message(self):
        """
        Cleanup the message
        :return:
        """

        message = string_cleanup(self.cleaned_data["message"])

        return message


class ReplyPersonalMessageForm(ModelForm):
    """
    Reply personal message form
    """

    message = forms.CharField(
        widget=CKEditorUploadingWidget(
            config_name="aa_forum",
            attrs={"rows": 10, "cols": 20, "style": "width: 100%;"},
        ),
        required=True,
        label=get_mandatory_form_label_text(_("Message")),
    )

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Meta definitions
        """

        model = PersonalMessage
        fields = ["message"]

    def clean_message(self):
        """
        Cleanup the message
        :return:
        """

        message = string_cleanup(self.cleaned_data["message"])

        return message



"""
Template tag

Do some useful stuff with a User object
"""

# Django
from django.contrib.auth.models import User
from django.template.defaulttags import register


@register.filter
def main_character_name(user: User) -> str:
    """
    Get the users main character name, or return the username if no main character
    :param user:
    :return:
    """

    if user is None:
        return ""

    try:
        return_value = user.profile.main_character.character_name
    except AttributeError:
        return str(user)

    return return_value


@register.filter
def main_character_id(user: User) -> int:
    """
    Get the users main character id, or return 1 if no main character
    :param user:
    :return:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.character_id
    except AttributeError:
        return_value = 1

    return return_value


@register.filter
def main_character_corporation_name(user: User) -> str:
    """
    Get the users main character corporation name,
    or an empty string if no main character
    :param user:
    :return:
    """

    if user is None:
        return ""

    try:
        return_value = user.profile.main_character.corporation_name
    except AttributeError:
        return_value = ""

    return return_value


@register.filter
def main_character_corporation_id(user: User) -> int:
    """
    Get the users main character corporation id, or 1 if no main character
    :param user:
    :return:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.corporation_id
    except AttributeError:
        return_value = 1

    return return_value


@register.filter
def main_character_alliance_name(user: User) -> str:
    """
    Get the users main character alliance name, or an empty string if no main character
    :param user:
    :return:
    """

    if user is None:
        return ""

    try:
        return_value = user.profile.main_character.alliance_name
    except AttributeError:
        return_value = ""

    return return_value


@register.filter
def main_character_alliance_id(user: User) -> int:
    """
    Get the users main character alliance id, or 1 if no main character
    :param user:
    :return:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.alliance_id

        # Check if the user is in an alliance
        try:
            return_value = int(return_value)
        except Exception:
            return_value = 1
    except AttributeError:
        return_value = 1

    return return_value



"""
Administration related views
"""

# Standard Library
import json

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import Group
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import Prefetch
from django.http import (
    HttpResponse,
    HttpResponseNotFound,
    HttpResponseRedirect,
    JsonResponse,
)
from django.shortcuts import redirect, render
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Forum
from aa_forum import __title__
from aa_forum.constants import DEFAULT_CATEGORY_AND_BOARD_SORT_ORDER
from aa_forum.forms import EditBoardForm, EditCategoryForm, NewCategoryForm, SettingForm
from aa_forum.helper.forms import message_form_errors
from aa_forum.models import Board, Category, Setting

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@permission_required("aa_forum.manage_forum")
def categories_and_boards(request: WSGIRequest) -> HttpResponse:
    """
    Administration index view
    :param request:
    :return:
    """

    categories = Category.objects.prefetch_related(
        Prefetch(
            "boards",
            queryset=Board.objects.filter(parent_board__isnull=True)
            .prefetch_related("groups")
            .prefetch_related("child_boards")
            .order_by("order", "id"),
        )
    ).order_by("order", "id")

    groups_queryset = Group.objects.all()
    category_loop = []

    for category in categories:
        boards_data = []

        for board in category.boards.all():
            child_boards_data = [
                {
                    "board_obj": child_board,
                    "board_forms": {
                        "board_edit_form": EditBoardForm(
                            prefix="edit-board-" + str(child_board.id),
                            instance=child_board,
                            groups_queryset=groups_queryset,
                        ),
                    },
                }
                for child_board in board.child_boards.all()
            ]

            boards_data.append(
                {
                    "board_obj": board,
                    "board_forms": {
                        "new_child_board_form": EditBoardForm(
                            prefix="new-child-board-" + str(board.id)
                        ),
                        "board_edit_form": EditBoardForm(
                            prefix="edit-board-" + str(board.id),
                            instance=board,
                            groups_queryset=groups_queryset,
                        ),
                    },
                    "child_boards": child_boards_data,
                }
            )

        category_data = {
            "category_obj": category,
            "category_forms": {
                "new_board": EditBoardForm(
                    prefix="new-board-in-category-" + str(category.id),
                    groups_queryset=groups_queryset,
                ),
                "edit_category": EditCategoryForm(
                    prefix="edit-category-" + str(category.id), instance=category
                ),
            },
            "boards": boards_data,
        }

        category_loop.append(category_data)

    form_new_category = NewCategoryForm(prefix="new-category")

    context = {
        "new_category_form": form_new_category,
        "category_loop": category_loop,
    }

    logger.info(f"{request.user} calling admin view.")

    return render(
        request, "aa_forum/view/administration/categories-and-boards.html", context
    )


@login_required
@permission_required("aa_forum.manage_forum")
def category_create(request: WSGIRequest) -> HttpResponseRedirect:
    """
    Create a new category
    :param request:
    :return:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = NewCategoryForm(request.POST, prefix="new-category")

        # Check whether it's valid:
        if form.is_valid():
            new_category = Category(
                name=form.cleaned_data["name"],
                order=DEFAULT_CATEGORY_AND_BOARD_SORT_ORDER,
            )
            new_category.save()

            # Add boards if any given
            if form.cleaned_data["boards"] != "":
                boards = form.cleaned_data["boards"]

                for board_name in boards.splitlines():
                    Board(
                        name=board_name,
                        category=new_category,
                        order=DEFAULT_CATEGORY_AND_BOARD_SORT_ORDER,
                    ).save()

            messages.success(
                request, mark_safe(_("<h4>Success!</h4><p>Category created.</p>"))
            )

            logger.info(f'{request.user} created category "{new_category.name}".')
        else:
            message_form_errors(request, form)

    return redirect("aa_forum:admin_categories_and_boards")


@login_required
@permission_required("aa_forum.manage_forum")
def category_edit(request: WSGIRequest, category_id: int) -> HttpResponseRedirect:
    """
    Edit a category
    :param request:
    :param category_id:
    :return:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditCategoryForm(
            request.POST, prefix="edit-category-" + str(category_id)
        )

        if form.is_valid():
            category = Category.objects.get(pk=category_id)
            category_name_old = category.name

            # Set new name
            category.name = form.cleaned_data["name"]
            category.save()

            messages.success(
                request,
                mark_safe(
                    _(
                        f'<h4>Success!</h4><p>Category name changed from "{category_name_old}" to "{category.name}".</p>'  # pylint: disable=line-too-long
                    )
                ),
            )

            logger.info(
                f"{request.user} changed category name "
                f'from "{category_name_old}" to "{category.name}".'
            )
        else:
            message_form_errors(request, form)

    return redirect("aa_forum:admin_categories_and_boards")


@login_required
@permission_required("aa_forum.manage_forum")
def category_delete(request: WSGIRequest, category_id: int) -> HttpResponseRedirect:
    """
    Edit a board
    :param request:
    :param category_id:
    :return:
    """

    try:
        category = Category.objects.get(pk=category_id)
    except Category.DoesNotExist:
        msg = f"Category with PK {category_id} not found."
        logger.warning(msg)
        return HttpResponseNotFound(msg)

    category_name = category.name
    category.delete()
    messages.success(
        request,
        mark_safe(_(f'<h4>Success!</h4><p>Category "{category_name}" removed.</p>')),
    )

    logger.info(f'{request.user} removed category "{category_name}".')

    return redirect("aa_forum:admin_categories_and_boards")


@login_required
@permission_required("aa_forum.manage_forum")
def board_create(request: WSGIRequest, category_id: int) -> HttpResponseRedirect:
    """
    Create a new board
    :param request:
    :param category_id:
    :return:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditBoardForm(
            request.POST, prefix="new-board-in-category-" + str(category_id)
        )

        board_category = Category.objects.get(pk=category_id)

        # Check whether it's valid:
        if form.is_valid():
            discord_webhook = (
                form.cleaned_data["discord_webhook"]
                if form.cleaned_data["discord_webhook"] != ""
                else None
            )

            new_board = Board(
                name=form.cleaned_data["name"],
                description=form.cleaned_data["description"],
                category=board_category,
                discord_webhook=discord_webhook,
                use_webhook_for_replies=form.cleaned_data["use_webhook_for_replies"],
                is_announcement_board=form.cleaned_data["is_announcement_board"],
                order=DEFAULT_CATEGORY_AND_BOARD_SORT_ORDER,
            )
            new_board.save()

            new_board.groups.set(form.cleaned_data["groups"])
            new_board.announcement_groups.set(form.cleaned_data["announcement_groups"])

            messages.success(
                request,
                mark_safe(
                    _(f'<h4>Success!</h4><p>Board "{new_board.name}" created.</p>')
                ),
            )

            logger.info(f'{request.user} created board "{new_board.name}".')
        else:
            message_form_errors(request, form)

    return redirect("aa_forum:admin_categories_and_boards")


@login_required
@permission_required("aa_forum.manage_forum")
def board_create_child(
    request: WSGIRequest, category_id: int, board_id: int
) -> HttpResponseRedirect:
    """
    Create a child board
    :param request:
    :param category_id:
    :param board_id:
    :return:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditBoardForm(request.POST, prefix="new-child-board-" + str(board_id))

        # Check whether it's valid:
        if form.is_valid():
            parent_board = Board.objects.get(pk=board_id)
            discord_webhook = (
                form.cleaned_data["discord_webhook"]
                if form.cleaned_data["discord_webhook"] != ""
                else None
            )

            new_board = Board(
                name=form.cleaned_data["name"],
                description=form.cleaned_data["description"],
                discord_webhook=discord_webhook,
                use_webhook_for_replies=form.cleaned_data["use_webhook_for_replies"],
                parent_board=parent_board,
                category=parent_board.category,
                order=DEFAULT_CATEGORY_AND_BOARD_SORT_ORDER,
            )
            new_board.save()

            messages.success(
                request,
                mark_safe(
                    _(f'<h4>Success!</h4><p>Board "{new_board.name}" created.</p>')
                ),
            )

            logger.info(
                f'{request.user} created board "{new_board.name}" '
                f'as child board of "{parent_board.name}".'
            )
        else:
            message_form_errors(request, form)

    return redirect("aa_forum:admin_categories_and_boards")


@login_required
@permission_required("aa_forum.manage_forum")
def board_edit(
    request: WSGIRequest, category_id: int, board_id: int
) -> HttpResponseRedirect:
    """
    Edit a board
    :param request:
    :param category_id:
    :param board_id:
    :return:
    """

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditBoardForm(request.POST, prefix="edit-board-" + str(board_id))

        # Check whether it's valid:
        if form.is_valid():
            discord_webhook = (
                form.cleaned_data["discord_webhook"]
                if form.cleaned_data["discord_webhook"] != ""
                else None
            )

            board = Board.objects.get(pk=board_id, category_id=category_id)
            board.name = form.cleaned_data["name"]
            board.description = form.cleaned_data["description"]
            board.discord_webhook = discord_webhook
            board.use_webhook_for_replies = form.cleaned_data["use_webhook_for_replies"]
            board.groups.set(form.cleaned_data["groups"])
            board.is_announcement_board = form.cleaned_data["is_announcement_board"]
            board.announcement_groups.set(form.cleaned_data["announcement_groups"])
            board.save()

            messages.success(
                request,
                mark_safe(_(f'<h4>Success!</h4><p>Board "{board.name}" changed.</p>')),
            )

            logger.info(f'{request.user} changed board "{board.name}".')
        else:
            message_form_errors(request, form)

    return redirect("aa_forum:admin_categories_and_boards")


@login_required
@permission_required("aa_forum.manage_forum")
def board_delete(
    request: WSGIRequest, category_id: int, board_id: int
) -> HttpResponseRedirect:
    """
    Delete a board
    :param request:
    :param category_id:
    :param board_id:
    :return:
    """

    try:
        board = Board.objects.get(pk=board_id)
    except Board.DoesNotExist:
        msg = f"Board with PK {board_id} not found."
        logger.warning(msg)

        return HttpResponseNotFound(msg)

    board_name = board.name
    board.delete()

    messages.success(
        request, mark_safe(_(f'<h4>Success!</h4><p>Board "{board_name}" removed.</p>'))
    )

    logger.info(f'{request.user} removed board "{board_name}".')

    return redirect("aa_forum:admin_categories_and_boards")


@login_required
@permission_required("aa_forum.manage_forum")
def ajax_category_order(request: WSGIRequest) -> JsonResponse:
    """
    Ajax call :: Save the category order
    :param request:
    :return:
    """

    data = []

    if request.method == "POST":
        categories = json.loads(request.POST.get("categories"))

        for category in categories:
            try:
                category_obj = Category.objects.get(pk=category["catId"])
            except Category.DoesNotExist:
                logger.warning(
                    "Tried to change order for non existing category with ID %s.",
                    category["catId"],
                )
            else:
                category_obj.order = category["catOrder"]
                category_obj.save(update_fields=["order"])

        data.append({"success": True})

    return JsonResponse(data, safe=False)


@login_required
@permission_required("aa_forum.manage_forum")
def ajax_board_order(request: WSGIRequest) -> JsonResponse:
    """
    Ajax call :: Save the board order
    :param request:
    :return:
    """

    data = []

    if request.method == "POST":
        boards = json.loads(request.POST.get("boards"))

        for board in boards:
            try:
                board_obj = Board.objects.get(pk=board["boardId"])
            except Board.DoesNotExist:
                logger.warning(
                    "Tried to change order for non existing board with ID %s.",
                    board["boardId"],
                )
            else:
                board_obj.order = board["boardOrder"]
                board_obj.save(update_fields=["order"])

        data.append({"success": True})

    return JsonResponse(data, safe=False)


@login_required
@permission_required("aa_forum.manage_forum")
def forum_settings(request: WSGIRequest) -> HttpResponse:
    """
    Forum Settings
    :param request:
    :return:
    """

    logger.info(f"{request.user} called forum settings page.")

    settings = Setting.objects.get(pk=1)

    # If this is a POST request we need to process the form data
    if request.method == "POST":
        settings_form = SettingForm(request.POST, instance=settings)

        # Check whether it's valid:
        if settings_form.is_valid():
            settings.messages_per_page = settings_form.cleaned_data["messages_per_page"]
            settings.topics_per_page = settings_form.cleaned_data["topics_per_page"]
            settings.user_signature_length = settings_form.cleaned_data[
                "user_signature_length"
            ]
            settings.save()

            messages.success(
                request, mark_safe(_("<h4>Success!</h4><p>Settings updated.<p>"))
            )

            return redirect("aa_forum:admin_forum_settings")
        else:
            messages.error(
                request,
                mark_safe(
                    _(
                        "<h4>Error!</h4>"
                        "<p>Something went wrong, please check your input.<p>"
                    )
                ),
            )
    else:
        settings_form = SettingForm(instance=settings)

    context = {"form": settings_form}

    return render(request, "aa_forum/view/administration/forum-settings.html", context)



"""
Set custom variables in a Django template
"""

# Django
from django import template
from django.contrib.auth.models import User
from django.utils.safestring import mark_safe

# AA Forum
from aa_forum.models import PersonalMessage

register = template.Library()


class SetVarNode(template.Node):
    """
    SetVarNode
    """

    def __init__(self, var_name, var_value):
        self.var_name = var_name
        self.var_value = var_value

    def render(self, context):
        """
        Render context
        :param context:
        :return:
        """

        try:
            value = template.Variable(self.var_value).resolve(context)
        except template.VariableDoesNotExist:
            value = ""

        context[self.var_name] = value

        return ""


@register.tag(name="set_template_variable")
def set_template_variable(parser, token):
    """
    Set a template variable
    Usage: {% set_template_variable <var_name> = <var_value> %}
    :param parser:
    :param token:
    :return:
    """

    parts = token.split_contents()

    if len(parts) < 4:
        raise template.TemplateSyntaxError(
            "'set_template_variable' tag must be of the form: "
            "{% set_template_variable <var_name> = <var_value> %}"
        )

    return SetVarNode(parts[1], parts[3])


@register.simple_tag
def personal_message_unread_count(user: User) -> str:
    """
    Return the number of new/unread personal messages
    :param user:
    :return:
    """

    return_value = ""
    message_count = PersonalMessage.objects.get_personal_message_unread_count_for_user(
        user
    )

    if message_count > 0:
        return_value = mark_safe(
            f'<span class="badge aa-forum-badge-personal-messages-unread-count">{message_count}</span>'  # pylint: disable=line-too-long
        )

    return return_value



# Generated by Django 4.0.7 on 2022-08-24 14:43

# Django
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("aa_forum", "0008_populate_default_settings"),
    ]

    operations = [
        migrations.AlterField(
            model_name="lastmessageseen",
            name="topic",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="last_message_seen",
                to="aa_forum.topic",
            ),
        ),
        migrations.AlterField(
            model_name="lastmessageseen",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="aa_forum_last_message_seen",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# Generated by Django 4.0.7 on 2022-08-24 09:08

# Django
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("aa_forum", "0006_reset_default_settings"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="setting",
            name="value",
        ),
        migrations.RemoveField(
            model_name="setting",
            name="variable",
        ),
        migrations.AddField(
            model_name="setting",
            name="default_max_messages",
            field=models.IntegerField(
                default=15,
                help_text="Maximum number of messages per page in the topic view",
                verbose_name="Messages per page",
            ),
        ),
        migrations.AddField(
            model_name="setting",
            name="default_max_topics",
            field=models.IntegerField(
                default=10,
                help_text="Maximum number of topics per page in the board view",
                verbose_name="Topics per page",
            ),
        ),
    ]



# Generated by Django 4.0.7 on 2022-08-30 20:17

# Django
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

# ckEditor
import ckeditor_uploader.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("aa_forum", "0011_userprofile"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="personalmessage",
            name="slug",
        ),
        migrations.AddField(
            model_name="personalmessage",
            name="deleted_by_recipient",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="personalmessage",
            name="deleted_by_sender",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="personalmessage",
            name="message_head",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="replies",
                to="aa_forum.personalmessage",
            ),
        ),
        migrations.AlterField(
            model_name="personalmessage",
            name="message",
            field=ckeditor_uploader.fields.RichTextUploadingField(),
        ),
        migrations.AlterField(
            model_name="personalmessage",
            name="recipient",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="personalmessage",
            name="sender",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



"""
Tests for the template tags
"""

# Django
from django.template import TemplateSyntaxError
from django.test import TestCase

# Alliance Auth
from allianceauth.tests.auth_utils import AuthUtils

# AA Forum
from aa_forum import __version__
from aa_forum.models import PersonalMessage, get_sentinel_user
from aa_forum.templatetags.aa_forum_template_variables import (
    personal_message_unread_count,
)
from aa_forum.tests.utils import create_fake_user, render_template


class TestMainCharacterName(TestCase):
    """
    Tests for main_character_name template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.template = "{% load aa_forum_user %}{{ user|main_character_name }}"

    def test_should_contain_character_name_for_users_with_main(self):
        """
        Test should contain character name for user with main set
        :return:
        """

        # given
        user = create_fake_user(1001, "Bruce Wayne")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "Bruce Wayne")

    def test_should_contain_user_character_name_for_users_without_main(self):
        """
        Should return username for users without a main character
        :return:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "john")

    def test_should_return_deleted_user_for_sentinel_user(self):
        """
        Should return "deleted" for sentinel user
        :return:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "deleted")

    def test_should_be_empty_for_none(self):
        """
        Test should be empty
        :return:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "")


class TestMainCharacterId(TestCase):
    """
    Tests for main_character_id template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.template = "{% load aa_forum_user %}{{ user|main_character_id }}"

    def test_should_contain_character_id_for_users_with_main(self):
        """
        Test should contain main character ID for users with main
        :return:
        """

        # given
        user = create_fake_user(1001, "Bruce Wayne")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1001")

    def test_should_contain_dummy_id_for_users_without_main(self):
        """
        Test should contain dummy ID (1) for users without main character
        :return:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")

    def test_should_be_dummy_id_for_sentinel_user(self):
        """
        Test should return dummy ID (1) for sentinel user
        :return:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")

    def test_should_be_dummy_id_for_none(self):
        """
        Test should return dummy ID (1) for None
        :return:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")


class TestMainCharacterCorporationName(TestCase):
    """
    Tests for main_character_corporation_name template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.template = (
            "{% load aa_forum_user %}{{ user|main_character_corporation_name }}"
        )

    def test_should_contain_corp_name_for_users_with_main(self):
        """
        Test should return corporation name for users with main character
        :return:
        """

        # given
        user = create_fake_user(1001, "Bruce Wayne", 2001, "Wayne Tech Inc.", "WYT")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "Wayne Tech Inc.")

    def test_should_be_empty_for_users_without_main(self):
        """
        Test should be empty for users without main character
        :return:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "")

    def test_should_be_empty_for_sentinel_user(self):
        """
        Test should be empty for sentinel user
        :return:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "")

    def test_should_be_empty_for_none(self):
        """
        Test should be empty for None
        :return:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "")


class TestMainCorporationId(TestCase):
    """
    Tests for main_character_corporation_id template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.template = (
            "{% load aa_forum_user %}{{ user|main_character_corporation_id }}"
        )

    def test_should_contain_corporation_id_for_users_with_main(self):
        """
        Test should return main character corp ID for users with main character
        :return:
        """

        # given
        user = create_fake_user(1001, "Bruce Wayne", 2001, "Wayne Tech Inc.", "WYT")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "2001")

    def test_should_be_dummy_id_for_users_without_main(self):
        """
        Test should return dummy ID (1) for users without main character
        :return:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)
        # then

        self.assertEqual(result, "1")

    def test_should_be_dummy_id_for_sentinel_user(self):
        """
        Test should return dummy ID (1) for sentinel user
        :return:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")

    def test_should_be_dummy_id_for_none(self):
        """
        Test should return dummy ID (1) for None
        :return:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")


class TestMainCharacterAllianceName(TestCase):
    """
    Tests for main_character_alliance_name template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.template = "{% load aa_forum_user %}{{ user|main_character_alliance_name }}"

    def test_should_contain_alliance_name_for_users_with_main(self):
        """
        Test should return main character alliance name for users with min character
        :return:
        """

        # given
        user = create_fake_user(
            1001,
            "Bruce Wayne",
            2001,
            "Wayne Tech Inc.",
            "WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprices",
        )

        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "Wayne Enterprices")

    def test_should_be_empty_for_users_without_main(self):
        """
        Test should be empty for users without main character
        :return:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "")

    def test_should_be_empty_when_main_is_not_in_an_alliance(self):
        """
        Test should be empty when main character is not in an alliance
        :return:
        """

        # given
        user = create_fake_user(
            2012, "William Riker", 2012, "Starfleet", "SF", alliance_id=None
        )

        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "")

    def test_should_be_empty_for_sentinel_user(self):
        """
        Test should be empty for sentinel user
        :return:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "")

    def test_should_be_empty_for_none(self):
        """
        Test should be empty for None
        :return:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "")


class TestMainAllianceId(TestCase):
    """
    Tests for main_character_alliance_id template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.template = "{% load aa_forum_user %}{{ user|main_character_alliance_id }}"

    def test_should_contain_alliance_id_for_users_with_main(self):
        """
        Test should return main character alliance ID for user with main character
        :return:
        """

        # given
        user = create_fake_user(
            1001,
            "Bruce Wayne",
            2001,
            "Wayne Tech Inc.",
            "WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "3001")

    def test_should_be_dummy_id_for_users_without_main(self):
        """
        Test should return dummy ID (1) for user without main character
        :return:
        """

        # given
        user = AuthUtils.create_user("john")
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")

    def test_should_dummy_id_when_main_is_not_in_an_alliance(self):
        """
        Test should dummy ID (1) when main character is not in an alliance
        :return:
        """

        # given
        user = create_fake_user(
            2012, "William Riker", 2012, "Starfleet", "SF", alliance_id=None
        )

        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")

    def test_should_be_dummy_id_for_sentinel_user(self):
        """
        Test should return dummy ID (1) for sentinel user
        :return:
        """

        # given
        user = get_sentinel_user()
        context = {"user": user}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")

    def test_should_be_dummy_id_for_none(self):
        """
        Test should return dummy ID (1) for None
        :return:
        """

        # given
        context = {"user": None}

        # when
        result = render_template(string=self.template, context=context)

        # then
        self.assertEqual(result, "1")


class TestForumTemplateVariables(TestCase):
    """
    Tests for aa_forum_template_variables template tag
    """

    def test_set_template_variable(self):
        """
        Test set template variable
        :return:
        """

        rendered = render_template(
            string=(
                "{% load aa_forum_template_variables %}"
                "{% set_template_variable foo = content %}"
                "{{ foo }}"
            ),
            context={"content": "bar"},
        )

        # self.assertInHTML("bar", rendered_template)
        self.assertEqual(rendered, "bar")

    def test_should_raise_template_syntax_error(self):
        """
        Test should raise template syntax error
        :return:
        """

        with self.assertRaisesMessage(
            TemplateSyntaxError,
            "'set_template_variable' tag must be of the form: "
            "{% set_template_variable <var_name> = <var_value> %}",
        ):
            render_template(
                "{% load aa_forum_template_variables %}"
                "{% set_template_variable foo %}"
                "{{ foo }}"
            )

    def test_should_return_personal_message_unread_count_as_empty_string(self):
        """
        Test personal_message_unread_count to return zero
        :return:
        """

        user = create_fake_user(
            1001,
            "Bruce Wayne",
            2001,
            "Wayne Tech Inc.",
            "WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )

        self.client.force_login(user)

        rendered = render_template(
            "{% load aa_forum_template_variables %}"
            "{% personal_message_unread_count 1001 %}"
        )

        self.assertEqual(rendered, "")

    def test_should_return_personal_message_unread_count_as_bootstrap_badge_with_number(
        self,
    ):
        """
        Test personal_message_unread_count to return a bootstrap badge with a number
        :return:
        """

        # given (creating our personal message)
        user_sender = create_fake_user(
            1001,
            "Bruce Wayne",
            2001,
            "Wayne Tech Inc.",
            "WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )

        user_receiver = create_fake_user(
            1002,
            "Batman",
            2001,
            "Wayne Tech Inc.",
            "WYT",
            alliance_id=3001,
            alliance_name="Wayne Enterprises",
        )

        PersonalMessage.objects.create(
            sender=user_sender,
            recipient=user_receiver,
            subject="Foo",
            message="Bar",
        )

        # when (template tage is triggered)
        response = personal_message_unread_count(user_receiver)

        # then
        self.assertEqual(
            response,
            '<span class="badge aa-forum-badge-personal-messages-unread-count">1</span>',
        )


class TestForumVersionedStatic(TestCase):
    """
    Tests for aa_forum_versioned_static template tag
    """

    def test_versioned_static(self):
        """
        Test should return static URL string with version
        :return:
        """

        context = {"version": __version__}

        rendered_template = render_template(
            string=(
                "{% load aa_forum_versioned_static %}"
                "{% aa_forum_static 'aa_forum/css/aa-forum.min.css' %}"
            ),
            context=context,
        )

        self.assertEqual(
            rendered_template,
            f'/static/aa_forum/css/aa-forum.min.css?v={context["version"]}',
        )


class TestHighlightSearchTerm(TestCase):
    """
    Tests for aa_forum_versioned_static template tag
    """

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        cls.message_text = "Lorem Ipsum"
        cls.message_with_link = '<a href="https://lorem-ipsum.com">Lorem Ipsum</a>'
        cls.message_with_link_without_href = '<a title="Lorem Ipsum">Lorem Ipsum</a>'
        cls.message_with_link_and_title = (
            '<a href="https://lorem-ipsum.com" title="Lorem Ipsum">Lorem Ipsum</a>'
        )
        cls.message_with_link_and_name = (
            '<a href="https://lorem-ipsum.com" name="Lorem Ipsum">Lorem Ipsum</a>'
        )
        cls.message_with_image = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg"> Lorem Ipsum'
        )
        cls.message_with_image_without_src = '<img alt="Lorem Ipsum"> Lorem Ipsum'
        cls.message_with_image_and_alt = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg" alt="Lorem Ipsum"> '
            "Lorem Ipsum"
        )
        cls.message_with_image_and_title = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg" title="Lorem Ipsum"> '
            "Lorem Ipsum"
        )
        cls.search_term = "Lorem"
        cls.template = (
            "{% load aa_forum_search %}"
            "{{ message|highlight_search_term:search_term  }}"
        )

    def test_should_highlight_with_just_text(self):
        """
        Test should highlight search term
        :return:
        """

        context = {"message": self.message_text, "search_term": self.search_term}

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(rendered_template, expected_result)

    def test_should_add_dummy_href(self):
        """
        Test should add a dummy href
        :return:
        """

        context = {
            "message": self.message_with_link_without_href,
            "search_term": self.search_term,
        }

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<a href="#" title="Lorem Ipsum">'
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum</a>'
        )

        self.assertEqual(rendered_template, expected_result)

    def test_should_not_highlight_in_link_href(self):
        """
        Test should not highlight in an a-tag's href attribute
        :return:
        """

        context = {"message": self.message_with_link, "search_term": self.search_term}

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<a href="https://lorem-ipsum.com">'
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum</a>'
        )

        self.assertEqual(rendered_template, expected_result)

    def test_should_not_highlight_in_link_title(self):
        """
        Test should not highlight in an a-tag's title attribute
        :return:
        """

        context = {
            "message": self.message_with_link_and_title,
            "search_term": self.search_term,
        }

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<a href="https://lorem-ipsum.com" title="Lorem Ipsum">'
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum</a>'
        )

        self.assertEqual(rendered_template, expected_result)

    def test_should_not_highlight_in_link_name(self):
        """
        Test should not highlight in an a-tag's name attribute
        :return:
        """

        context = {
            "message": self.message_with_link_and_name,
            "search_term": self.search_term,
        }

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<a href="https://lorem-ipsum.com" name="Lorem Ipsum">'
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum</a>'
        )

        self.assertEqual(rendered_template, expected_result)

    def test_should_not_highlight_in_img_src(self):
        """
        Test should not highlight in an img-tag's src attribute
        :return:
        """

        context = {
            "message": self.message_with_image,
            "search_term": self.search_term,
        }

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg"/> '
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(rendered_template, expected_result)

    def test_should_add_dummy_image_src(self):
        """
        Test should add a dummy image src
        :return:
        """

        context = {
            "message": self.message_with_image_without_src,
            "search_term": self.search_term,
        }

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<img alt="Lorem Ipsum" src="#"/> '
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(rendered_template, expected_result)

    def test_should_not_highlight_in_img_alt(self):
        """
        Test should not highlight in an img-tag's alt attribute
        :return:
        """

        context = {
            "message": self.message_with_image_and_alt,
            "search_term": self.search_term,
        }

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<img alt="Lorem Ipsum" src="https://lorem-ipsum.com/lorem-ipsum.jpg"/> '
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(rendered_template, expected_result)

    def test_should_not_highlight_in_img_title(self):
        """
        Test should not highlight in an img-tag's title attribute
        :return:
        """

        context = {
            "message": self.message_with_image_and_title,
            "search_term": self.search_term,
        }

        rendered_template = render_template(
            string=self.template,
            context=context,
        )

        expected_result = (
            '<img src="https://lorem-ipsum.com/lorem-ipsum.jpg" title="Lorem Ipsum"/> '
            '<span class="aa-forum-search-term-highlight">Lorem</span> Ipsum'
        )

        self.assertEqual(rendered_template, expected_result)



"""
Our app setting
"""


# Django
from django.apps import apps

# pylint: disable=import-outside-toplevel unused-import


def discordproxy_installed() -> bool:
    """
    Check if discordproxy is installed by trying to import its DiscordClient
    :return:
    """

    try:
        # Third Party
        from discordproxy.client import DiscordClient  # noqa: F401
    except ModuleNotFoundError:
        return False
    else:
        return True


def allianceauth_discordbot_installed() -> bool:
    """
    Check if allianceauth-discordbot is installed and active
    :return:
    """

    return apps.is_installed("aadiscordbot")


def discord_messaging_proxy_available() -> bool:
    """
    Check if any known discord messaging proxy is available
    :return:
    """

    return discordproxy_installed() or allianceauth_discordbot_installed()
